import math
import os
import random

# from numpy import array, float32
from fsgs.platform import PlatformHandler
from arcade.glui.constants import TOP_ITEM_ARROW, TOP_ITEM_LEFT
from arcade.glui.constants import ROW_NAME, ROW_PLATFORM

from fsbc.util import memoize
from fsgs.Database import Database
from fsgs.util.gamenameutil import GameNameUtil
from arcade.resources import resources, gettext
from .font import BitmapFont
from .opengl import gl, fs_emu_blending, fs_emu_texturing
from .render import Render
from arcade.glui.state import State
from arcade.glui.texture import Texture
from arcade.glui.texturemanager import TextureManager
from .errordialog import show_exception

LIGHTING = False
DIRTY_WHILE_NOT_LOADED = True


def create_item_menu(name):
    from arcade.glui.itemmenu import ItemMenu

    return ItemMenu(name)


def create_mesh(width, height, upsidedown):
    hw = width / 2
    hh = height / 2
    if LIGHTING:
        res = 200
        radius = 0.018
    else:
        res = 2
        radius = 0.0
    # radius = 0.1
    mesh = []
    for y in range(0, res + 1):
        y = -hh + y * height / res
        line = []
        mesh.append(line)
        for x in range(0, res + 1):
            x = -hw + x * width / res
            # d = min(x, y, res- x, res -y) / (res / 2.0)
            # rx = x * 2.0 / res
            z = radius
            nx = 0.0
            ny = 0.0
            nz = 1.0
            # edge = False
            # using max to account for rounding errors - d cannot be negative
            d = max(0, hw - abs(x))
            # print(hw, x, d)
            if d < radius:
                d = radius - d
                # d = radius - 1.0 - x
                # print(d, d / radius)
                a = math.acos(d / radius)
                # print(a)
                z = math.sin(a) * radius
                nx = -d if x < 0 else d
                nz = z
                # edge = True
            # using max to account for rounding errors - d cannot be negative
            d = max(0, hh - abs(y))
            if d < radius:
                d = radius - d
                # d = radius - 1.0 - x
                # print(d, d / radius)
                a = math.acos(d / radius)
                # print(a)
                z = min(z, math.sin(a) * radius)
                ny = -d if y < 0 else d
                nz = z
                # edge = True
            nl = math.sqrt(nx ** 2 + ny ** 2 + nz ** 2)
            nx /= nl
            ny /= nl
            nz /= nl
            line.append((x, y, z, nx, ny, nz))
            # if edge:
            #     line.append((x, y, -z, nx, ny, -nz))
    return mesh


# @memoize
# def get_vbo_for_cover(width, height):
#     floats = []
#     upsidedown=False
#     mesh = create_mesh(width, height, upsidedown)
#     hw = width / 2
#     hh = height / 2
#     mesh_size = len(mesh)
#     print(mesh_size)
#     count = 0
#     for iy in range(mesh_size - 1):
#         for ix in range(mesh_size - 1):
#             def vertex(ix, iy):
#                 x, y, z, nx, ny, nz = mesh[iy][ix]
#                 if upsidedown:
#                     s, t = (x + hw) / width, (y + hh) / height
#                 else:
#                     s, t = (x + hw) / width, 1.0 - (y + hh) / height
#                 floats.extend([s, t, nx, ny, nz, x, y, z])
#             vertex(ix, iy)
#             vertex(ix + 1, iy)
#             vertex(ix + 1, iy + 1)
#             vertex(ix, iy + 1)
#             count += 4
#     # FIXME: specify usage with usage kw arg (specify as "static" data,
#     # used often)
#     vbo = VBO(array(floats, dtype=float32))
#     return vbo, count


def render_cover(texture, width, height):
    if not texture:
        texture = Texture.missing_cover
    texture.bind()
    x = -width / 2.0
    # y = height / 2.0 - 0.5 - height / 2.0
    y = -0.5
    # if reflection:

    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(0.33, 0.33, 0.33)
    gl.glTexCoord2f(0.0, 1.0)
    gl.glVertex2f(x, y)
    gl.glTexCoord2f(1.0, 1.0)
    gl.glVertex2f(x + width, y)
    gl.glTexCoord2f(1.0, 0.0)
    gl.glVertex2f(x + width, y - height)
    gl.glTexCoord2f(0.0, 0.0)
    gl.glVertex2f(x, y - height)
    gl.glEnd()

    # else:
    gl.glBegin(gl.GL_QUADS)
    gl.glColor3f(1.0, 1.0, 1.0)
    gl.glTexCoord2f(0.0, 1.0)
    gl.glVertex2f(x, y)
    gl.glTexCoord2f(1.0, 1.0)
    gl.glVertex2f(x + width, y)
    gl.glTexCoord2f(1.0, 0.0)
    gl.glVertex2f(x + width, y + height)
    gl.glTexCoord2f(0.0, 0.0)
    gl.glVertex2f(x, y + height)
    gl.glEnd()

    # glPushMatrix()
    # glTranslatef(0.0, height / 2.0 - 0.5, 0.0)
    # vbo, count = get_vbo_for_cover(width, height)
    # vbo.bind()
    # glInterleavedArrays(GL_T2F_N3F_V3F, 0, None)
    # if reflection:
    #     glScalef(1.0, -1.0, 1.0)
    #     glTranslatef(0.0, height, 0.0)
    #     if LIGHTING:
    #         # No point in drawing everything
    #         glDrawArrays(GL_QUADS, 0, count // 4)
    #     else:
    #         glDrawArrays(GL_QUADS, 0, count)
    # else:
    #     glDrawArrays(GL_QUADS, 0, count)
    # vbo.unbind()
    # glPopMatrix()


class MenuItem(object):
    def __init__(self, title=""):
        self.ratio = 0.75
        self.title = title
        self.supertitle = ""
        self.subtitle = ""
        self.path_title = title
        self.path_title_active = ""
        # x, y, w and h are used by top menu items
        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.enabled = True

    def activate(self, menu):
        print("Activate not overidden")

    @memoize
    def get_image_files(self):
        path = self.get_image_path()
        if path:
            return [path]
        return []

    def get_default_texture(self):
        return Texture.default_item

    def get_texture(self):
        path = self.get_image_path()
        if path:
            try:
                texture = TextureManager.get().get_texture(path)
            except KeyError:
                texture = None
            if texture:
                if texture is Texture.default_item:
                    return Texture.missing_cover
                return texture
            else:
                Render.get().dirty = DIRTY_WHILE_NOT_LOADED
        return Texture.missing_cover
        # return self.get_default_texture()

    @memoize
    def get_image_path(self):
        name = self.get_resource_name()
        if name:
            try:
                return resources.resource_filename(
                    os.path.join(u"items", name + u".png")
                )
            except Exception:
                pass
        return None

    def get_resource_name(self):
        return None

    def render(
        self,
        width,
        height,
        border_color=(0.0, 0.0, 0.0, 1.0),
        inner_border_color=(0.92, 0.92, 0.92),
        height_offset=0.0,
        brightness=1.0,
        ratio=None,
        area=None,
    ):
        render_cover(self.get_texture(), width, height)

    def create_menu_path(self, menu):
        path = []
        # while menu:
        if not menu:
            return path
        path.append(menu.selected_item)
        for p in reversed(menu.parents):
            # path.insert(0, menu.selected_item)
            path.insert(0, p.selected_item)
            # menu = menu.parent_menu
        print("create_menu_path", path)
        for p in path:
            print(p.path_title)
        return path

    def get_category_filter(self):
        return None

    def get_game_filter(self):
        return None

    def generate_category_filters(self, menu_path):
        for item in menu_path:
            flt = item.get_category_filter()
            if flt:
                yield flt

    def generate_game_filters(self, menu_path):
        for item in menu_path:
            flt = item.get_game_filter()
            if flt:
                yield flt

    def generate_category_items(self, menu_path):
        category_filters = list(self.generate_category_filters(menu_path))

        def check(cat):
            for flt in category_filters:
                if not flt(cat):
                    return False
            return True

        categories = [
            PlatformMenuItem(),
            LetterMenuItem(),
            ShuffleMenuItem(),
            ListMenuItem(),
            # YearMenuItem(),
            # KeywordMenuItem(),
        ]
        if len(menu_path) == 0:
            categories.insert(0, AllMenuItem())
        return [item for item in categories if check(item)]

    def generate_game_items(self, menu_path):
        args = []
        clause = []
        list_uuid = None
        for item in menu_path:
            c, a = item.get_filter_clause()
            if c:
                clause.append(c)
                args.extend(a)
            if item.get_filter_list_uuid():
                list_uuid = item.get_filter_list_uuid()
        # clause = " ".join(clause)

        filters = list(self.generate_game_filters(menu_path))
        print("filters for", menu_path, "-", filters)
        return self.create_game_items(clause, args, filters, list_uuid)

    @classmethod
    def create_game_items(cls, words, args=None, filters=None, list_uuid=None):
        if filters is None:
            filters = []
        search = " ".join(words)
        item_list = []
        local_game_database = Database.instance()
        for game in local_game_database.find_games_new(
            search=search, database_only=True, list_uuid=list_uuid
        ):
            item = GameItem(game)
            for filter in filters:
                pass
                # if not filter(game):
                #     break
            else:
                item_list.append(item)
        return item_list

    def get_filter_clause(self):
        return "", []

    def get_filter_list_uuid(self):
        return None

    def update_size_left(self):
        self.update_size(self.get_top_left_text())

    def update_size_right(self):
        self.update_size(self.get_top_right_text())

    def update_size(self, text):
        # w, h = Render.get().measure_text(text, Font.main_path_font)
        w, h = BitmapFont.title_font.measure(text)
        self.w = w + 20 + 20

    def get_top_left_text(self):
        return self.path_title_active.upper() or self.path_title.upper()

    def get_top_right_text(self):
        return self.title.upper()

    def render_top_background(
        self,
        selected,
        style=TOP_ITEM_ARROW,
        mouse_state=False,
        mouse_pressed_state=False,
    ):
        x, y, w, h = self.x, self.y, self.w, self.h
        z = -0.01 - 0.01 * x / 1920
        selected = selected or mouse_state
        if selected:
            fs_emu_texturing(True)
            fs_emu_blending(True)
            if mouse_pressed_state:
                alpha = 0.75
            else:
                alpha = 1.0
            Texture.top_item_background.render(x, y, w, h, z, opacity=alpha)
            # fs_emu_texturing(False)
            # fs_emu_blending(False)
            # gl.glBegin(gl.GL_QUADS)
            # gl.glColor3f(0.00, 0x99 / 0xff, 0xcc / 0xff)
            # gl.glVertex3f(x + 4, y + 4, z)
            # gl.glVertex3f(x + w - 4, y + 4, z)
            # gl.glVertex3f(x + w - 4, y + h - 4, z)
            # gl.glVertex3f(x + 4, y + h - 4, z)
            # gl.glEnd()

            # fs_emu_blending(True)
            # glColor3f(1.0, 1.0, 1.0)
            #
            # #z = -0.01
            #
            # if style == TOP_ITEM_LEFT:
            #     if selected:
            #         texture = Texture.top_item_left_selected
            #     else:
            #         texture = Texture.top_item_left
            #     texture.bind()
            #     tw, th = texture.size
            #     w = 60
            #
            #     glBegin(GL_QUADS)
            #     glTexCoord2f(0.0, 1.0)
            #     glVertex3f(x, y, z)
            #     glTexCoord2f(1.0, 1.0)
            #     glVertex3f(x + w, y, z)
            #     glTexCoord2f(1.0, 0.0)
            #     glVertex3f(x + w, y + h, z)
            #     glTexCoord2f(0.0, 0.0)
            #     glVertex3f(x, y + h, z)
            #     glEnd()
            #     x += 60
            #
            # w = self.w - 60
            #
            # if selected:
            #     texture = Texture.top_item_selected
            #     x = x - 60
            #     w = w + 60
            # else:
            #     texture = Texture.top_item
            # texture.bind()
            # tw, th = texture.size
            # glBegin(GL_QUADS)
            # glTexCoord2f(0.0, 1.0)
            # glVertex3f(x, y, z)
            # glTexCoord2f(1.0, 1.0)
            # glVertex3f(x + w, y, z)
            # glTexCoord2f(1.0, 0.0)
            # glVertex3f(x + w, y + h, z)
            # glTexCoord2f(0.0, 0.0)
            # glVertex3f(x, y + h, z)
            # glEnd()
            #
            # if style == TOP_ITEM_ARROW or style == TOP_ITEM_RIGHT:
            #     if style == TOP_ITEM_ARROW:
            #         if selected:
            #             texture = Texture.top_item_arrow_selected
            #         else:
            #             texture = Texture.top_item_arrow
            #         texture.bind()
            #         tw, th = texture.size
            #     else:
            #         Texture.top_item_right.bind()
            #         tw, th = Texture.top_item_right.size
            #     x = x + w
            #     w = 60
            #
            #     glBegin(GL_QUADS)
            #     glTexCoord2f(0.0, 1.0)
            #     glVertex3f(x, y, z)
            #     glTexCoord2f(1.0, 1.0)
            #     glVertex3f(x + w, y, z)
            #     glTexCoord2f(1.0, 0.0)
            #     glVertex3f(x + w, y + h, z)
            #     glTexCoord2f(0.0, 0.0)
            #     glVertex3f(x, y + h, z)
            #     glEnd()
            #
            # fs_emu_blending(False)

    def render_top_left(self, selected=False):
        state = State.get()
        mouse_state = state.mouse_item == self
        mouse_pressed_state = mouse_state and state.mouse_press_item == self
        self.render_top_background(
            selected,
            style=TOP_ITEM_LEFT,
            mouse_state=mouse_state,
            mouse_pressed_state=mouse_pressed_state,
        )
        text = self.get_top_left_text()
        self.render_top(text, selected)

    def render_top_right(self, selected=False):
        state = State.get()
        mouse_state = state.mouse_item == self
        mouse_pressed_state = mouse_state and state.mouse_press_item == self
        self.render_top_background(
            selected,
            style=TOP_ITEM_LEFT,
            mouse_state=mouse_state,
            mouse_pressed_state=mouse_pressed_state,
        )
        self.render_top(self.get_top_right_text(), selected)

    def render_top(self, text="", selected=False, right_align=False):
        if right_align:
            tw, th = BitmapFont.title_font.measure(text)
            x = self.x + self.w - 20 - tw
        else:
            x = self.x + 20
            # if selected:
            #     color = (0.0, 0.0, 0.0)
            # else:
            #     color = (0.85, 0.85, 0.85)
            # tw, th = Render.get().text(text, Font.main_path_font,
            #         x, self.y + 30, h=TOP_HEIGHT_VISIBLE, color=color)
        BitmapFont.title_font.render(text, x, self.y + 14)
        # print(BitmappedFont.title_font.h)

    def get_screen_texture(self, n):
        return Texture.missing_screenshot
        # return None


class AutoExpandItem(MenuItem):
    def activate(self, menu):
        new_menu = create_item_menu(self.title)
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        want_selected = len(new_menu)
        for item in self.generate_game_items(menu_path):
            new_menu.append(item)
        if len(new_menu) > want_selected:
            new_menu.set_selected_index(want_selected, immediate=True)
        new_menu.add_add_item()
        return new_menu


class DummyItem(MenuItem):
    def __init__(self, title):
        MenuItem.__init__(self)
        self.title = title


class NoItem(MenuItem):
    def __init__(self, title="N/A"):
        MenuItem.__init__(self)
        self.title = title

    def get_default_texture(self):
        return Texture.missing_cover

    def activate(self, menu):
        return


class NoLastPlayedItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("No Last Played")


class ListMenuItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("List")
        self.path_title_active = "Choose List"

    def get_resource_name(self):
        return "list"

    def activate(self, menu):
        print("ListMenuItem.activate")
        new_menu = create_item_menu(gettext("Select List"))
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        # Database.cursor.execute("SELECT DISTINCT platform FROM Game "
        #         "ORDER BY platform")
        # items = []
        # for row in Database.cursor:
        #     item = PlatformItem(row[0])
        #     items.append((item.sort_title, item))
        # for sort_title, item in sorted(items):
        #    new_menu.append(item)

        for title, path in ListItem.get_game_lists():
            item = ListItem(title, path)
            new_menu.append(item)
        return new_menu


class ListItem(AutoExpandItem):
    def __init__(self, title, path):
        AutoExpandItem.__init__(self)
        self.list_path = path
        self.title = title
        # self.platform = platform
        self.path_title = title
        self.subtitle = ""
        self.sort_title = self.title
        # self.game_filter_set = set()
        self._list = None

    @classmethod
    def get_game_lists(cls):
        result = []
        local_game_database = Database.instance()
        for list_uuid, list_name in local_game_database.get_game_lists():
            result.append((list_name, list_uuid))
        return result

    @classmethod
    def get_favorites_uuid(cls):
        for list_name, list_uuid in cls.get_game_lists():
            # FIXME: Should use some common function to determine this
            if list_name == "Favorites":
                return list_uuid
        raise LookupError("No favorites list")

    def get_category_filter(self):
        def category_filter(category):
            if category.__class__ == ListMenuItem:
                return False
            return True

        return category_filter

    def get_filter_list_uuid(self):
        return self.list_path

        # @memoize

    # def get_list_contents(self):
    #     if self._list is None:
    #         self._list = set()
    #         local_game_database = Database.instance()
    #         local_game_database.get
    #     return self._list

    # def get_game_filter(self):
    #
    #     def game_filter(game):
    #         game_uuids = set()
    #         if game.uuid in game_uuids:
    #             return True
    #         return False
    #
    #     return game_filter

    # def get_filter_clause(self):
    #     # FIXME: NUMBERS
    #     return "AND platform = ?", [self._platform]
    #     print(name)


class PlatformMenuItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("Platform")
        self.path_title_active = "Choose Platform"

    def get_resource_name(self):
        return "platform"

    def activate(self, menu):
        print("PlatformMenuItem.activate")
        new_menu = create_item_menu(gettext("Select Platform"))
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        with Database.instance() as database:
            cursor = database.cursor()
            cursor.execute(
                "SELECT DISTINCT platform FROM game WHERE have >= 3 "
                "ORDER BY platform"
            )
            items = []
            for row in cursor:
                if not row[0]:
                    continue
                item = PlatformItem(row[0])
                items.append((item.sort_title, item))
            for sort_title, item in sorted(items):
                new_menu.append(item)
        return new_menu


class PlatformItem(AutoExpandItem):
    def __init__(self, platform):
        AutoExpandItem.__init__(self)
        self._platform = platform
        # using self.platform for "description"

        # self.sort_title = ""
        # try:
        #    self.title, self.subtitle, self.sort_title = {
        #        "Amiga": (
        #            "Amiga",
        #            "Commodore Amiga",
        #            "Commodore 2"),
        #        "amstrad-cpc": (
        #            "Amstrad CPC",
        #            "Color Personal Computer",
        #            ""),
        #        "apple-II": (
        #            "Apple II",
        #            "II/Plus/e/c",
        #            ""),
        #        "arcade": (
        #            "Arcade Video Games",
        #            "Multiple Arcade Machines",
        #            ""),
        #        "atari-2600": (
        #            "Atari 2600",
        #            "Video Computer System",
        #            "Atari 1"),
        #        "atari-8-bit": (
        #            "Atari 8-Bit",
        #            "Atari 400/800/XL/XE",
        #            "Atari 2"),
        #        "atari-5200": (
        #            "Atari 5200",
        #            "Atari 5200 SuperSystem",
        #            "Atari 3"),
        #        "atari-7800": (
        #            "Atari 7800",
        #            "Atari 7800 ProSystem",
        #            "Atari 4"),
        #        "atari-st": (
        #            "Atari ST",
        #            "Atari ST/STE",
        #            "Atari 5"),
        #        "bbc-micro": (
        #            "BBC Micro",
        #            "Acorn BCC Micro",
        #            ""),
        #        "commodore-64": (
        #            "Commodore 64",
        #            "",
        #            "Commodore 1"),
        #        "game-boy": (
        #            "Game Boy",
        #            "",
        #            "Nintendo 2"),
        #        "game-boy-color": (
        #            "Game Boy Color",
        #            "",
        #            "Nintendo 5"),
        #        "game-boy-advance": (
        #            "Game Boy Advance",
        #            "",
        #            "Nintendo 6"),
        #        "game-cube": (
        #            "GameCube",
        #            "Nintendo GameCube",
        #            "Nintendo 7"),
        #        "game-gear": (
        #            "Game Gear",
        #            "Sega Game Gear",
        #            "Sega 3"),
        #        "lynx": (
        #            "Atari Lynx",
        #            "",
        #            "Atari 6"),
        #        "nintendo": (
        #            "Nintendo",
        #            "",
        #            "Nintendo 1"),
        #        "nintendo-64": (
        #            "Nintendo 64",
        #            "",
        #            "Nintendo 4"),
        #        "DOS": (
        #            "DOS",
        #            "IBM PC Compatible",
        #            "PC 1"),
        #        "master-system": (
        #            "Master System",
        #            "Sega Master System / Mark III",
        #            "Sega 1"),
        #        "mega-drive": (
        #            "Mega Drive",
        #            "Sega Mega Drive / Genesis",
        #            "Sega 2"),
        #        "playstation": (
        #            "PlayStation",
        #            "Sony PlayStation",
        #            ""),
        #        "playstation-2": (
        #            "PlayStation 2",
        #            "Sony PlayStation 2",
        #            ""),
        #        "super-nintendo": (
        #            "Super Nintendo",
        #            "Entertainment System",
        #            "Nintendo 3"),
        #        "turbografx-16": (
        #            "TurboGrafx-16",
        #            "Entertainment SuperSystem",
        #            ""),
        #        "wii": (
        #            "Wii",
        #            "Nintendo Wii",
        #            "Nintendo 8"),
        #        "windows": (
        #            "Windows",
        #            "IBM PC Compatible",
        #            "PC 2"),
        #        "zx-spectrum": (
        #            "ZX Spectrum",
        #            "Sinclair ZX Spectrum",
        #            ""),
        #    }[platform]
        # except KeyError:
        #    self.title = platform
        self.platform = platform
        self.path_title = platform

        self.title = PlatformHandler.get_platform_name(platform)
        self.path_title = self.title

        # self.title = ""
        # self.title = "CHOOSE PLATFORM: " + self.title
        # self.subtitle = ""
        self.subtitle = "CHOOSE PLATFORM"

        # if not self.sort_title:
        #     self.sort_title = self.title

        # FIXME: for now, test using just title as sort title
        self.sort_title = self.title

    def get_category_filter(self):
        def category_filter(category):
            if category.__class__ == PlatformMenuItem:
                return False
            return True

        return category_filter

    # def get_game_filter(self):
    #     def game_filter(game_info):
    #         if game_info.platform == self._platform:
    #             return True
    #         return False
    #     return game_filter

    @memoize
    def get_resource_name(self):
        name = ""
        for c in self._platform.lower():
            if c in "abcdefghijklmnopqrstuvwxyz0123456789":
                name = name + c
        return name

    def get_filter_clause(self):
        return "s:{0}".format(self._platform), []


class AllMenuItem(AutoExpandItem):
    def __init__(self):
        AutoExpandItem.__init__(self)
        self.title = gettext("All")
        self.path_title = "All Games"

    def get_resource_name(self):
        return "all"

    def get_category_filter(self):
        def category_filter(category):
            return False

        return category_filter

        # def get_game_filter(self):
        #     def game_filter(game_info):
        #         return True
        #     return game_filter


class ShuffleMenuItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("Shuffle")
        self.path_title = "Shuffle"
        # FIXME: Sort game list in random order

    def get_resource_name(self):
        return "shuffle"

    def activate(self, menu):
        new_menu = create_item_menu(self.title)
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        game_items = list(self.generate_game_items(menu_path))
        random.shuffle(game_items)
        for item in game_items:
            new_menu.append(item)
        return new_menu


class LetterMenuItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("Letter")
        self.path_title_active = "Choose Letter"

    def get_resource_name(self):
        return "letter"

    def activate(self, menu):
        print("LetterMenuItem.activate")
        new_menu = create_item_menu(gettext("Select First Letter"))
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        for letter in "#ABCDEFGHIJKLMNOPQRSTUVWXYZ":
            item = LetterItem(letter)
            new_menu.append(item)
        return new_menu


class LetterItem(AutoExpandItem):
    def __init__(self, letter):
        AutoExpandItem.__init__(self)
        self.letter = letter
        self.title = letter
        self.path_title = letter

    def get_resource_name(self):
        if self.letter == "#":
            return "0"
        return self.letter.lower()

    def get_category_filter(self):
        def category_filter(category):
            if category.__class__ == LetterMenuItem:
                return False
            return True

        return category_filter

    # def get_game_filter(self):
    #     def game_filter(game_info):
    #         if game_info.sort_title[0] == self.letter:
    #             return True
    #         if self.letter == "#":
    #             if game_info.sort_title[0] in "0123456789":
    #                 return True
    #         return False
    #     return game_filter

    def get_filter_clause(self):
        # FIXME: NUMBERS
        # if self.letter == "#":
        #     clause = ["AND (0 = 1"]
        #     params = []
        #     for c in "0123456789":
        #         clause.append(" OR name LIKE ?")
        #         params.append("{0}%".format(c))
        #     clause.append(")")
        #     return "".join(clause), params
        # return "AND name LIKE ?", ["{0}%".format(self.letter)]
        return "l:{0}".format(self.letter.lower()), []


class YearMenuItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("Year")
        self.path_title_active = "Choose Year"

    def activate(self, menu):
        print("YearMenuItem.activate")
        menu = create_item_menu(gettext("Select Year"))
        # for year in GameList().get().get_years(): # FIXME: GET YEARS FROM
        # FILTERED GAME LIST
        #     item = YearItem(year)
        #     menu.append(item)
        return menu


class YearItem(AutoExpandItem):
    def __init__(self, year):
        AutoExpandItem.__init__(self)
        self.year = year
        self.title = str(year)
        self.path_title = str(year)

    def get_category_filter(self):
        def category_filter(category):
            if category.__class__ == YearMenuItem:
                return False
            return True

        return category_filter

        # def get_game_filter(self):
        #     def game_filter(game_info):
        #         if game_info.year == self.year:
        #             return True
        #         return False
        #     return game_filter


class KeywordMenuItem(MenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = gettext("Keyword")
        self.path_title_active = "Choose Keyword"

    def activate(self, menu):
        print("KeywordMenuItem.activate")
        new_menu = create_item_menu(gettext("Select Keyword"))
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        category_filters = list(self.generate_category_filters(menu_path))

        def check(cat):
            for flt in category_filters:
                if not flt(cat):
                    return False
            return True

        # for keyword in GameList().get().get_keywords():
        #     item = KeywordItem(keyword)
        #     if check(item):
        #         new_menu.append(item)
        return new_menu


class KeywordItem(AutoExpandItem):
    def __init__(self, keyword):
        AutoExpandItem.__init__(self)
        self.keyword = keyword
        self.title = keyword
        self.path_title = keyword

    def get_category_filter(self):
        def category_filter(category):
            if category.__class__ == KeywordItem:
                if category.keyword == self.keyword:
                    return False
            return True

        return category_filter

        # def get_game_filter(self):
        #     def game_filter(game_info):
        #         if self.keyword in game_info.keywords:
        #             return True
        #         return False
        #     return game_filter


class GameItem(MenuItem):
    def __init__(self, game_info):
        MenuItem.__init__(self)
        self.game_info = game_info
        # self.id = id
        self.uuid = game_info[0]
        self.title = game_info[1]

        # id, name, configurations, platform, ratio=0.75,
        #     screenratio=1.33, year="", publisher="", developer="",
        #     subtitlepos=-1

        # when a game item is the selected item in the parent menu,
        # we are displaying category items
        self.path_title_active = "Add Filter"

        # self.name = name
        self.configurations = []
        # self.platform = platform
        # print("self.platform = ", self.platform)
        # self.title = name
        # FIXME: Get correct year
        # self.year = year
        # self.publisher = publisher
        # self.developer = developer
        self.subtitle = ""

        ratio = 0.75
        screenratio = 1.33

        if ratio:
            self.ratio = ratio
        else:
            # if self.platform == "Game Boy":
            #     self.ratio = 1.0
            # elif self.platform == "Game Boy Color":
            #     self.ratio = 1.0
            # elif self.platform == "Game Boy Advance":
            #     self.ratio = 1.0
            # elif self.platform == "Super Nintendo":
            #     self.ratio = 1.33
            # elif self.platform == "Nintendo 64":
            #     self.ratio = 1.33
            # else:
            #     self.ratio = 0.75
            self.ratio = 0.75
        self.screenratio = screenratio or 1.33

        # FIXME: REMOVE
        self.ratio = 0.75

        # if subtitlepos > 0:
        #     self.title = name[:subtitlepos].strip()
        #     self.subtitle = name[subtitlepos:].strip()
        # elif subtitlepos < 0:
        #     self.supertitle = name[:-subtitlepos].strip()
        #     self.title = name[-subtitlepos:].strip()
        # else:
        #     self.fix_titles()

    @property
    def name(self):
        return self.game_info[1]

    # @property
    # def title(self):
    #     return self.game_info[1]

    @property
    def platform(self):
        return self.game_info[2]

    @property
    def year(self):
        return self.game_info[3]

    @property
    def publisher(self):
        return self.game_info[4]

    def fix_titles(self):
        try:
            title, subtitle = self.title.split(":", 1)
        except ValueError:
            pass
        else:
            # if len(title) >= 8:
            self.title = title
            self.subtitle = subtitle
        if not self.subtitle:
            if "(" in self.title:
                if self.title[-1] == ")":
                    pos = self.title.index("(")
                    self.subtitle = self.title[pos + 1 : -1].strip()
                    self.title = self.title[:pos].strip()
        # if not self.subtitle:
        #     if len(self.title) > 30 and " in " in self.title:
        #         pos = self.title.find(" in ")
        #         if pos > 8:
        #             self.subtitle = self.title[pos+4:].strip()
        #             self.title = self.title[:pos+4].strip()
        #             self.subtitle = self.subtitle[0].upper() +
        # self.subtitle[1:]
        if not self.subtitle:
            try:
                title, subtitle = self.title.split(" - ", 1)
            except ValueError:
                pass
            else:
                if len(title) >= 8:
                    self.title = title
                    self.subtitle = subtitle
        if not self.subtitle:
            pos = self.title.find(" and the ")
            if pos > 7:
                self.subtitle = self.title[pos:].strip()
                self.title = self.title[:pos].strip()
                # self.subtitle = self.subtitle[0].upper() + self.subtitle[1:]

    def activate(self, menu):
        try:
            from arcade.glui.gamemenu import GameMenu

            new_menu = GameMenu(self)
        except Exception:
            show_exception()
            print("returning None")
            return None
        return new_menu

    def get_default_texture(self):
        return Texture.missing_cover

    @memoize
    def get_image_path(self):
        cmp_name = GameNameUtil.create_cmpname(self.name)

        # FIXME: FIXME: DISABLED TEMPORARILY
        # FIXME: USING 512-size only for now

        # if False and os.path.exists("../fs-game-database/game/texture"):
        # if os.path.exists("../fs-game-database/game/texture"):
        #     base_path = "../fs-game-database/game/texture"
        #     path = os.path.join(base_path, NameUtil.create_cmpname(
        #             self.platform), cmp_name, "front-512.jpg")
        #             #self.platform), cmp_name, "front-1024.jpg")
        #     if os.path.exists(path):
        #         return path
        # #else:
        # base_path = os.path.join(pyapp.app.get_data_dir(), "texture")
        # if os.path.exists("c:\\data\\game\\info"):
        #     base_path = "c:\\data\\game\\info"
        # else:
        #     base_path = os.path.join(pyapp.app.get_data_dir(), "info")
        # base_path = os.path.join(GameCenter.data_dir, "info")
        # path = os.path.join(base_path, NameUtil.create_cmpname(self.platform),
        #         cmp_name, "front.jpg")
        # if not os.path.exists(path):
        #     base_path = "c:\\data\\game\\info"
        #     path = os.path.join(base_path, NameUtil.create_cmpname(
        #             self.platform), cmp_name, "front.jpg")
        #         #cmp_name, "front-1024.jpg")
        # if not os.path.exists(path):
        #     path = os.path.join(base_path,
        #             NameUtil.create_cmpname(self.platform),
        #             cmp_name, "front-512.jpg")
        # path = os.path.join(NameUtil.create_cmpname(self.platform),
        #         cmp_name, "front.jpg")
        # return path
        # return self.game_info[5] + "?s=512&t=jpg"
        if not self.game_info[5]:
            return None
        return self.game_info[5] + "?w=480&h=640&t=cc&f=jpg"

    # @memoize
    # def get_image_files(self):
    #     path = self.get_image_path()
    #     if os.path.exists(path):
    #         return [path]
    #     return []

    @memoize
    def get_image_files(self):
        path = self.get_image_path()
        # if path and os.path.exists(path):
        if path:
            return [path]
        return []

    @memoize
    def get_screen_path(self, n):
        # cmp_name = NameUtil.create_cmpname(self.name)
        # base_path = os.path.join(GameCenter.data_dir, "info")
        # path = os.path.join(base_path, NameUtil.create_cmpname(self.platform),
        #         cmp_name, "screen{0}.png".format(n))
        # if not os.path.exists(path):
        #     base_path = "c:\\data\\game\\info"
        #     path = os.path.join(base_path, NameUtil.create_cmpname(
        #             self.platform), cmp_name, "screen{0}.png".format(n))
        # if not os.path.exists(path):
        #     return None

        assert 1 <= n <= 5
        value = self.game_info[7 + n - 1]
        # if value:
        #     value = "sha1:" + value

        # """
        # if n == 1:
        #     name = "screen1.png"
        # elif n == 2:
        #     name = "title.png"
        # else:
        #     name = "screen{0}.png".format(n - 1)
        # """
        # name = "screen{0}.png".format(n)
        # path = os.path.join(NameUtil.create_cmpname(
        #         self.platform), cmp_name, name)
        if not value:
            return None
        return value + "?s=1x"

    def get_screen_texture(self, n):
        path = self.get_screen_path(n)
        # print("get_screen_texture", repr(path))

        if path:
            # FIXME: LOAD WITH THE REST OF THE COVERS!
            TextureManager.get().load_images([path])
            try:
                texture = TextureManager.get().get_texture(path)
            except KeyError:
                texture = None
            if not texture:
                Render.get().dirty = DIRTY_WHILE_NOT_LOADED
            if texture is Texture.default_item:
                return Texture.missing_screenshot
            return texture
        return Texture.missing_screenshot



import weakref

from arcade.glui.bezier import Bezier
from arcade.glui.state import State


class AnimationSystem(object):
    animation_list = []

    @classmethod
    def is_active(cls):
        return len(cls.animation_list) > 0

    @classmethod
    def update(cls, t=None):
        t = t or State.get().time
        deleted = []
        for ref_obj in cls.animation_list:
            obj = ref_obj()
            if obj is None:
                deleted.append(ref_obj)
            else:
                if not obj.update(t):
                    deleted.append(ref_obj)
        for d in deleted:
            cls.animation_list.remove(d)


class AnimateValueLinear(object):
    def __init__(
        self, what, from_value, from_time, to_value, to_time, run=True
    ):
        self.what = what
        self.from_value = from_value
        self.to_value = to_value
        self.from_time = from_time
        self.to_time = to_time
        if run:
            AnimationSystem.animation_list.append(weakref.ref(self))

    @property
    def max_time(self):
        return self.to_time

    def update(self, t=None):
        t = t or State.get().time
        progress = (t - self.from_time) / (self.to_time - self.from_time)
        progress = max(0.0, min(1.0, progress))
        setattr(
            self.what[0],
            self.what[1],
            self.from_value + progress * (self.to_value - self.from_value),
        )
        return t < self.to_time


class AnimateValueBezier(object):
    def __init__(
        self, what, x0, t0, x1, t1, x2, t2, x3, t3, points=50, run=True
    ):
        self.what = what
        self.params = (x0, t0, x1, t1, x2, t2, x3, t3)
        self.curve = Bezier.bezier(
            (t0, x0), (t1, x1), (t2, x2), (t3, x3), steps=points
        )
        if run:
            AnimationSystem.animation_list.append(weakref.ref(self))

    @property
    def max_time(self):
        return self.curve[-1][0]

    def update(self, t=None):
        x0, t0, x1, t1, x2, t2, x3, t3 = self.params
        t = t or State.get().time
        if t < t0:
            t = t0
        elif t > t3:
            t = t3
        x = Bezier.interpolate(self.curve, t)
        setattr(self.what[0], self.what[1], x)
        return t < t3


class AnimationSequence(object):
    def __init__(self, anim_seq):
        self.anim_seq = anim_seq
        AnimationSystem.animation_list.append(weakref.ref(self))

    def update(self, t=None):
        t = t or State.get().time
        for anim in self.anim_seq:
            if t > anim.max_time:
                continue
            anim.update(t)
            break
        else:
            return self.anim_seq[-1].update()
        return True



import fsui
from launcher.ui.widgets import CloseButton
from .skin import LauncherTheme


class PagedDialog(fsui.Window):
    page_changed = fsui.Signal()

    def __init__(self, parent, title):
        super().__init__(parent, title, minimizable=False, maximizable=False)
        self.theme = LauncherTheme.get()
        self.layout = fsui.VerticalLayout()

        # self.layout.set_padding(20)
        layout = self.layout

        h_layout = fsui.HorizontalLayout()
        layout.add(h_layout, fill=True, expand=True)

        layout_2 = fsui.VerticalLayout()
        # layout_2.padding_top = 20
        # layout_2.padding_bottom = 20
        # layout_2.padding_left = 20

        self.list_view = fsui.ListView(self, border=False)
        self.list_view.set_min_width(240)
        self.list_view.item_selected.connect(self.on_select_item)

        self.list_view.set_row_height(
            self.window.theme.sidebar_list_row_height
        )
        self.list_view.setStyleSheet(
            """
        QListView {{
            padding-top: 20px;
            background-color: {base};
            outline: none;
        }}
        QListView::item {{
            padding-left: 20px;
            border: 0px;
        }}
        QListView::item:selected {{
            background-color: {row_bg};
            color: {row_fg};
        }}
        """.format(
                row_fg=self.window.theme.sidebar_list_row_text.to_hex(),
                row_bg=self.window.theme.sidebar_list_row_background.to_hex(),
                base=self.window.theme.sidebar_list_background.to_hex(),
            )
        )
        layout_2.add(self.list_view, fill=True, expand=True)
        h_layout.add(layout_2, fill=True)

        # h_layout.add_spacer(20)

        v_layout = fsui.VerticalLayout()
        h_layout.add(v_layout, fill=True, expand=True)

        use_scrolling = False
        if use_scrolling:
            # FIXME: Problems with page height
            scroll_area = fsui.VerticalScrollArea(self)
            self.page_container = fsui.Panel(scroll_area)
            self.page_container.layout = fsui.VerticalLayout()
            scroll_area.set_widget(self.page_container)
            v_layout.add(scroll_area, fill=True, expand=True)
        else:
            self.page_container = fsui.Panel(self)
            self.page_container.layout = fsui.VerticalLayout()
            v_layout.add(self.page_container, fill=True, expand=True)

        # self.layout.add_spacer(20)

        # h_layout = fsui.HorizontalLayout()
        # layout.add(h_layout, fill=True)

        self.page_titles = []
        self.pages = []

        # self.add_page(_("Hardware"), HardwarePage)
        # self.add_page(_("Hard Drives"), HardDrivesPage)
        # elf.add_page(_("Custom Options"), CustomOptionsPage)

        self.button_layout = fsui.HorizontalLayout()
        v_layout.add(self.button_layout, fill=True, margin=20, margin_top=0)
        if self.window.theme.has_close_buttons:
            self.button_layout.add_spacer(expand=True)
            self.close_button = CloseButton(self)
            self.button_layout.add(
                self.close_button, fill=True, margin_left=10
            )

        self.current_page = None

        # self.page_container.set_min_width(600)
        # self.page_container.set_min_height(640)
        # self.set_size_from_layout()
        self.set_size((840, 640))
        self.center_on_parent()

    # def on_close_button(self):
    #     self.end_modal(0)

    def on_select_item(self, index):
        self.set_page(index)

    @property
    def index(self):
        return self.list_view.get_index()

    @property
    def page(self):
        return self.current_page

    def get_index(self):
        return self.list_view.get_index()

    def get_page_title(self, index):
        # return self.page_titles[index][0]
        return self.page_titles[index]["label"]

    def get_page_index_by_title(self, title):
        # for index, t in enumerate(self.page_titles):
        #     if t[0] == title:
        for i, item in enumerate(self.page_titles):
            if item["label"] == title:
                return i
        return None

    def add_page(self, title, function, icon=None, bold=False):
        self.page_titles.append({"label": title, "icon": icon, "bold": bold})
        self.pages.append(function)
        self.list_view.set_items(self.page_titles)

    def set_page(self, index):
        if self.current_page:
            self.current_page.hide()
            self.page_container.layout.remove(self.current_page)
        if callable(self.pages[index]):
            page = self.pages[index](self.page_container)
            self.pages[index] = page
        else:
            page = self.pages[index]
        self.page_container.layout.add(page, fill=True, expand=True)
        self.current_page = page
        page.show()
        # print("calling self.layout.update")
        self.page_container.layout.update()
        self.layout.update()

        self.page_changed.emit()

    def set_page_by_title(self, title):
        index = self.get_page_index_by_title(title)
        if index is not None:
            # self.set_page(index)
            self.list_view.set_index(index)
        return index



import sys

import fsui
from fsgs.platform import Platform
from launcher.option import Option
from fsgs.context import fsgs
from launcher.devicemanager import DeviceManager
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_signal import LauncherSignal
from launcher.ui.HelpButton import HelpButton
from launcher.ui.IconButton import IconButton
from launcher.ui.behaviors.platformbehavior import (
    AMIGA_PLATFORMS,
    AmigaShowBehavior,
)

MIN_TYPE_CHOICE_WIDTH = 200


class InputSelector(fsui.Group):
    def __init__(self, parent, port, autofire_button=True):
        self.port = port
        self.device_option_key = "joystick_port_{0}".format(port)
        self.mode_option_key = "joystick_port_{0}_mode".format(port)
        self.autofire_mode_option_key = "joystick_port_{0}_autofire".format(
            port
        )

        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        if port == 1:
            non_amiga_port_gui_index = 0
        elif port == 0:
            non_amiga_port_gui_index = 1
        else:
            non_amiga_port_gui_index = port
        self.layout.add(
            InputPortTypeChoice(self, fsgs, non_amiga_port_gui_index)
        )

        self.joystick_mode_values = [
            "nothing",
            "joystick",
        ]
        self.joystick_mode_titles = [
            gettext("No Amiga Device"),
            gettext("Amiga Joystick"),
        ]

        if port < 2:
            self.joystick_mode_values.extend([
                "mouse",
                "cd32 gamepad",
            ])
            self.joystick_mode_titles.extend([
                gettext("Amiga Mouse"),
                gettext("CD32 Pad"),
            ])

        self.mode_choice = fsui.Choice(self, self.joystick_mode_titles)
        self.mode_choice.set_min_width(MIN_TYPE_CHOICE_WIDTH)
        # AmigaEnableBehavior(self.mode_choice)
        AmigaShowBehavior(self.mode_choice)

        self.layout.add(self.mode_choice)
        self.layout.add_spacer(10)
        # else:
        #     self.mode_choice = None
        if port >= 4:
            self.mode_choice.disable()

        # devices = ["", _("No Host Device"), _("Mouse"),
        #         _("Cursor Keys and Right Ctrl/Alt")]
        # for i, name in enumerate(DeviceManager.get_joystick_names()):
        #     devices.append(name)
        #     if not self.joystick_values_initialized:
        #         self.joystick_values.append(DeviceManager.device_ids[i])
        # self.joystick_values_initialized = True

        self.device_choice = fsui.ComboBox(self, [""], read_only=True)
        # AmigaEnableBehavior(self.device_choice)
        self.joystick_values = []
        self.rebuild_device_list()
        self.device_choice.set_index(0)
        # print(self.device_choice.get_index())
        AmigaShowBehavior(self.device_choice)
        self.layout.add(self.device_choice, expand=True)

        self.layout.add(
            InputPortDeviceChoice(self, fsgs, non_amiga_port_gui_index),
            expand=True,
        )

        if port < 4 and autofire_button:
            self.autofire_button = IconButton(self, "16x16/lightning_off.png")
            self.autofire_button.activated.connect(self.on_autofire_button)
            self.layout.add(self.autofire_button, margin_left=10)
            AmigaShowBehavior(self.autofire_button)
        else:
            self.autofire_button = None

        if port == 4:
            self.help_button = HelpButton(
                self, "https://fs-uae.net/custom-joystick-port"
            )
            self.layout.add(self.help_button, margin_left=10)

        self.initialize_from_config()
        self.set_config_handlers()

    def rebuild_device_list(self):
        self.joystick_values = ["", "none"]
        devices = ["", gettext("No Host Device")]
        for i, name in enumerate(DeviceManager.get_joystick_names()):
            devices.append(fix_device_name(name))
            self.joystick_values.append(DeviceManager.device_ids[i])
        self.device_choice.set_items(devices)

    def initialize_from_config(self):
        self.on_config(
            self.device_option_key, LauncherConfig.get(self.device_option_key)
        )
        self.on_config(
            self.mode_option_key, LauncherConfig.get(self.mode_option_key)
        )
        self.on_config(
            self.autofire_mode_option_key,
            LauncherConfig.get(self.autofire_mode_option_key),
        )

    def set_config_handlers(self):
        if self.mode_choice is not None:
            self.mode_choice.on_changed = self.on_mode_changed
        self.device_choice.on_changed = self.on_device_changed
        LauncherConfig.add_listener(self)
        LauncherSignal.add_listener("settings_updated", self)
        LauncherSignal.add_listener("device_list_updated", self)

    def on_destroy(self):
        print("on_destroy")
        LauncherConfig.remove_listener(self)
        LauncherSignal.remove_listener("settings_updated", self)
        LauncherSignal.remove_listener("device_list_updated", self)

    def on_mode_changed(self):
        if self.mode_choice is not None:
            index = self.mode_choice.get_index()
            value = self.joystick_mode_values[index]
            self.set_value_or_default(value)

    def set_value_or_default(self, value):
        if self.port == 0:
            if value == "mouse":
                value = ""
        elif self.port == 1:
            if LauncherConfig.get("amiga_model").startswith("CD32"):
                default = "cd32 gamepad"
            else:
                default = "joystick"
            if value == default:
                value = ""
        else:
            if value == "nothing":
                value = ""
        if LauncherConfig.get(self.mode_option_key) != value:
            LauncherConfig.set(self.mode_option_key, value)

    def on_device_changed(self):
        index = self.device_choice.get_index()

        value = self.joystick_values[index]
        if value != "none":
            # Reset to default device for other ports using the same device.
            for port in range(4):
                if self.port == port:
                    continue
                key = "joystick_port_{0}".format(port)
                if LauncherConfig.get(key) == value:
                    LauncherConfig.set(key, "")
        LauncherConfig.set(self.device_option_key, value)

    def on_autofire_button(self):
        if LauncherConfig.get(self.autofire_mode_option_key) == "1":
            LauncherConfig.set(self.autofire_mode_option_key, "")
        else:
            LauncherConfig.set(self.autofire_mode_option_key, "1")

    def on_config(self, key, value):
        if key == "platform":
            self.layout.update()
            return

        if key == "amiga_model":
            value = LauncherConfig.get(
                "joystick_port_{0}_mode".format(self.port)
            )
            self.set_value_or_default(value)

        if key == self.mode_option_key or key == "amiga_model":
            value = DeviceManager.get_calculated_port_mode(
                LauncherConfig, self.port
            )
            for i, config in enumerate(self.joystick_mode_values):
                if config == value:
                    if self.mode_choice is not None:
                        self.mode_choice.set_index(i)
                        if self.port >= 4:
                            self.device_choice.enable(i != 0)
                    break
            else:
                print("FIXME: could not set mode")
        elif key == self.device_option_key or key == "amiga_model":
            # print(self.joystick_values)
            value_lower = value.lower()
            for i, name in enumerate(self.joystick_values):
                if value_lower == name.lower():
                    self.device_choice.set_index(i)
                    break
        elif key == self.autofire_mode_option_key:
            if self.autofire_button is not None:
                if value == "1":
                    self.autofire_button.set_tooltip(
                        gettext("Auto-Fire is On")
                    )
                    self.autofire_button.set_icon_name(
                        "16x16/lightning_red.png"
                    )
                else:
                    self.autofire_button.set_tooltip(
                        gettext("Auto-Fire is Off")
                    )
                    self.autofire_button.set_icon_name(
                        "16x16/lightning_off.png"
                    )

        # this is intended to catch all config changes for all ports (both
        # mode and device) to update the defaults
        if key.startswith("joystick_port_") or key == "amiga_model":
            self.update_default_device()

    def on_device_list_updated_signal(self):
        # print(self.device_choice.get_index())
        had_default = self.device_choice.get_index() == 0
        self.rebuild_device_list()
        self.update_default_device(had_default=had_default)

    def on_settings_updated_signal(self):
        self.update_default_device()

    def update_default_device(self, had_default=None):
        config = {}
        for port in range(4):
            key = "joystick_port_{0}".format(port)
            if self.port == port:
                config[key] = ""
            else:
                config[key] = LauncherConfig.get(key)
            key = "joystick_port_{0}_mode".format(port)
            config[key] = DeviceManager.get_calculated_port_mode(
                LauncherConfig, port
            )
        device = DeviceManager.get_device_for_port(config, self.port)
        default_description = gettext("Default ({0})").format(
            fix_device_name(device.name)
        )
        # print("default_description = ", default_description)

        if had_default is None:
            had_default = self.device_choice.get_index() == 0
        # print("had default", had_default, self.device_choice.get_index())
        self.device_choice.set_item_text(0, default_description)
        # print("had_default", had_default)
        if had_default:
            # print("set text for", self.port, default_description)
            # self.device_choice.set_index(1)
            self.device_choice.set_text(default_description)
            self.device_choice.set_index(0)
            # print(self.device_choice.get_index())


class InputPortTypeChoice(fsui.Choice):
    def __init__(self, parent, fsgc, port_gui_index):
        self.fsgc = fsgc
        self._choice_values = []
        self._choice_labels = []
        self.port_gui_index = port_gui_index
        self.port = self.port_gui_index + 1
        super().__init__(parent, self._choice_labels)
        self._platform = ""
        self._config_key = ""
        self.fsgc.signal.connect("config", self.on_config)
        self.on_config(Option.PLATFORM, self.fsgc.config.get(Option.PLATFORM))
        self.changed.connect(self.__changed)
        self.set_min_width(MIN_TYPE_CHOICE_WIDTH)

    def on_destroy(self):
        self.fsgc.signal.disconnect("config", self.on_config)

    def __changed(self):
        self.fsgc.config.set(
            self._config_key, self._choice_values[self.get_index()]
        )

    def on_config(self, key, value):
        if key == Option.PLATFORM:
            self.port = self.port_gui_index + 1
            if value == Platform.C64:
                if self.port_gui_index == 0:
                    self.port = 2
                elif self.port_gui_index == 1:
                    self.port = 1
            self._platform = value
            self._config_key = "{}_port_{}_type".format(value, self.port)
            self.update_options()
            self.update_index(self.fsgc.config.get(self._config_key))
            self.update_enabled()
        elif key == self._config_key:
            self.update_index(value)

    def update_enabled(self):
        self.set_visible(self._platform not in AMIGA_PLATFORMS)
        self.set_enabled(self._choice_labels != ["N/A"])

    def update_index(self, value):
        try:
            index = self._choice_values.index(value)
        except ValueError:
            index = 0
        with self.changed.inhibit:
            self.set_index(index)

    def update_options(self):
        try:
            option = Option.get(self._config_key)
        except KeyError:
            self._choice_values = ["0"]
            self._choice_labels = ["N/A"]
        else:
            choices = option["values"]
            self._choice_values = [x[0] for x in choices]
            self._choice_labels = [x[1] for x in choices]

            if "default" in option:
                for x in choices:
                    if x[0] == option["default"]:
                        default_label = x[1] + " (*)"
                        break
                else:
                    default_label = "??? (*)"
                self._choice_values.insert(0, "")
                self._choice_labels.insert(0, default_label)
        with self.changed.inhibit:
            self.clear()
            for label in self._choice_labels:
                self.add_item(label)


class InputPortDeviceChoice(fsui.ComboBox):
    def __init__(self, parent, fsgc, port_gui_index):
        super().__init__(parent, [""], read_only=True)
        self.fsgc = fsgc
        self.port_gui_index = port_gui_index
        self.port = self.port_gui_index + 1
        self._platform = ""
        self._config_key = ""
        self.device_option_key = ""

        # AmigaEnableBehavior(self.device_choice)
        self.device_values = []
        self.rebuild_device_list()

        self.fsgc.signal.connect("config", self.on_config)
        # Must check platform before device option key
        self.on_config(Option.PLATFORM, self.fsgc.config.get(Option.PLATFORM))
        self.on_config(
            self.device_option_key, LauncherConfig.get(self.device_option_key)
        )
        # self.changed.connect(self.__changed)
        self.set_index(0)

        LauncherConfig.add_listener(self)
        LauncherSignal.add_listener("settings_updated", self)
        LauncherSignal.add_listener("device_list_updated", self)

    def rebuild_device_list(self):
        self.device_values = ["", "none"]
        devices = ["", gettext("No Host Device")]
        for i, name in enumerate(DeviceManager.get_joystick_names()):
            devices.append(fix_device_name(name))
            self.device_values.append(DeviceManager.device_ids[i])
        self.set_items(devices)

    def on_destroy(self):
        print("on_destroy")
        LauncherConfig.remove_listener(self)
        LauncherSignal.remove_listener("settings_updated", self)
        LauncherSignal.remove_listener("device_list_updated", self)

    def on_changed(self):
        index = self.get_index()
        value = self.device_values[index]
        if value != "none":
            # Reset to default device for other ports using the same device.
            for port in range(1, 4 + 1):
                if self.port == port:
                    continue
                key = "{}_port_{}".format(self._platform, port)
                if LauncherConfig.get(key) == value:
                    LauncherConfig.set(key, "")
        LauncherConfig.set(self.device_option_key, value)

    def update_enabled(self):
        self.set_visible(self._platform not in AMIGA_PLATFORMS)
        # self.set_enabled(self._choice_labels != ["N/A"])

    def on_config(self, key, value):
        if key == "platform":
            self.port = self.port_gui_index + 1
            if value == Platform.C64:
                if self.port_gui_index == 0:
                    self.port = 2
                elif self.port_gui_index == 1:
                    self.port = 1
            self._platform = value
            self.update_enabled()
            self.device_option_key = "{}_port_{}".format(
                self._platform, self.port
            )
            # Disable the control if the type option does not exist
            try:
                Option.get("{}_port_{}_type".format(self._platform, self.port))
            except KeyError:
                self.disable()
            else:
                self.enable()

            return
        # if key == self.mode_option_key or key == "amiga_model":
        #     value = DeviceManager.get_calculated_port_mode(
        #         LauncherConfig, self.port)
        #     for i, config in enumerate(self.joystick_mode_values):
        #         if config == value:
        #             if self.mode_choice is not None:
        #                 self.mode_choice.set_index(i)
        #                 if self.port >= 4:
        #                     self.enable(i != 0)
        #             break
        #     else:
        #         print("FIXME: could not set mode")
        # elif key == self.type_option_key:
        elif key == self.device_option_key:
            value_lower = value.lower()
            for i, name in enumerate(self.device_values):
                if value_lower == name.lower():
                    self.set_index(i)
                    break

        # This is intended to catch all config changes for all ports (both
        # mode and device) to update the defaults

        # FIXME:
        # if key.startswith("joystick_port_") or key == "amiga_model":
        #     self.update_default_device()

        if key.startswith("{}_port_".format(self._platform)):
            self.update_default_device()

    def on_device_list_updated_signal(self):
        # print(self.get_index())
        had_default = self.get_index() == 0
        self.rebuild_device_list()
        self.update_default_device(had_default=had_default)

    def on_settings_updated_signal(self):
        self.update_default_device()

    def update_default_device(self, had_default=None):
        config = {"platform": self._platform}
        for port in range(1, 4 + 1):
            key = "{}_port_{}".format(self._platform, port)
            if self.port == port:
                config[key] = ""
            else:
                config[key] = LauncherConfig.get(key)
            key = "{}_port_{}_type".format(self._platform, port)
            config[key] = LauncherConfig.get(key)
            # config[key] = DeviceManager.get_calculated_port_mode(
            #     LauncherConfig, port)

        device = DeviceManager.get_non_amiga_device_for_port(config, self.port)

        default_description = "{} (*)".format(fix_device_name(device.name))
        # print("default_description = ", default_description)

        if had_default is None:
            had_default = self.get_index() == 0
        # print("had default", had_default, self.get_index())
        self.set_item_text(0, default_description)
        # print("had_default", had_default)
        if had_default:
            # print("set text for", self.port, default_description)
            # self.set_index(1)
            self.set_text(default_description)
            self.set_index(0)
            # print(self.get_index())


def fix_device_name(name):
    system = gettext("System")
    if sys.platform == "win32":
        system = "Windows"
    elif sys.platform.startswith("linux"):
        system = "Linux"
    elif sys.platform == "darwin":
        system = "OS X"

    if name == "Keyboard":
        return gettext("{system} Keyboard").format(system=system)
    elif name == "Mouse":
        return gettext("Mouse: {system} Mouse").format(system=system)
    else:
        return gettext(name)



import fsui
from .canvas import Canvas
from .painter import Painter


class Button(fsui.Button):
    pass


class CustomButton(Canvas):

    activated = fsui.Signal

    def __init__(self, parent):
        super().__init__(parent)

        self.hover = False
        self.pressed = False

    def on_activate(self):
        pass

    def on_left_down(self):
        self.pressed = True
        # For menu popups
        self.on_mouse_motion()
        self.refresh()

    def on_mouse_enter(self):
        if not self.hover:
            self.hover = True
            self.refresh()

    def on_mouse_leave(self):
        # Mouse leave event is received when a menu is popped up, so we
        # explicitly check for mouse position.
        self.on_mouse_motion()

    def on_mouse_motion(self):
        new_state = self.is_mouse_over()
        if self.hover != new_state:
            self.hover = new_state
            self.refresh()

    def on_left_up(self):
        self.pressed = False
        self.refresh()
        if self.is_mouse_over():
            self.on_activate()

    def on_left_dclick(self):
        self.on_left_down()


class FlatButton(CustomButton):
    def __init__(self, parent):
        super().__init__(parent)
        # self.set_background_color(fsui.Color(0xff, 0xff, 0xff))
        self.set_min_width(40)
        self.set_min_height(34)
        self.border_colors = [
            fsui.Color(0xBC, 0xBC, 0xBC),
            fsui.Color(0xAD, 0xAD, 0xAD),
            fsui.Color(0x8C, 0x8C, 0x8C),
        ]
        self.fill_colors = [
            fsui.Color(0xFF, 0xFF, 0xFF),
            fsui.Color(0xE6, 0xE6, 0xE6),
            fsui.Color(0xD4, 0xD4, 0xD4),
        ]

    def on_paint(self):
        p = Painter(self)
        size = self.get_size()
        if self.hover and self.pressed:
            border_color = self.border_colors[2]
            fill_color = self.fill_colors[2]
        elif self.hover:
            border_color = self.border_colors[1]
            fill_color = self.fill_colors[1]
        else:
            border_color = self.border_colors[0]
            fill_color = self.fill_colors[0]
        # p.draw_rectangle(0, 0, size[0], size[1], border_color)
        # p.draw_rectangle(1, 1, size[0] - 2, size[1] - 2, fill_color)
        p.rounded_rectangle(
            0,
            0,
            size[0],
            size[1],
            fill_color,
            6,
            border_color=border_color,
            border_width=1.0,
        )



import logging
import os
from typing import List, Dict, DefaultDict, BinaryIO

from fsbc import settings
from fsgs.archive import Archive
from fsgs.option import Option


def should_disable_drive_click() -> bool:
    """Check if drive clicks should be disabled when generating a config."""
    return True


def create_prefs_file(config: DefaultDict[str, str], path: str) -> bool:
    """Write a WHDLoad.prefs respecting options from config.

    Returns true if the prefs file was created.
    """
    if config[Option.NETPLAY_GAME]:
        # The options below are commonly retrieved from settings, not
        # config, and settings are not synced in net play, so we use
        # default settings.
        logging.info("[WHDLOAD] No WHDLoad overrides in net play mode")
        return False
    prefs = default_whdload_prefs
    splash_delay = config[Option.WHDLOAD_SPLASH_DELAY]
    if splash_delay:
        prefs = prefs.replace(
            ";SplashDelay=0", "SplashDelay={}".format(int(splash_delay))
        )

    quit_key = config[Option.WHDLOAD_QUIT_KEY]
    if quit_key:
        prefs = prefs.replace(
            ";QuitKey=$45", "QuitKey=${}".format(
                quit_key.upper()))

    # Make sure the data is CRLF line terminated.
    prefs = prefs.replace("\r\n", "\n").replace("\n", "\r\n")
    with open(path, "wb") as f:
        f.write(prefs.encode("ISO-8859-1"))
    return True


def override_config(config: DefaultDict[str, str]):
    if should_disable_drive_click():
        config[Option.FLOPPY_DRIVE_VOLUME_EMPTY] = "0"
    model = settings.get(Option.WHDLOAD_MODEL)
    if model:
        if model == "A1200":
            config[Option.AMIGA_MODEL] = "A1200"
        elif model == "A1200/NONCE":
            # The following slaves do not work with A1200 non-cycle-exact
            # (non-exhaustive list, only some random tests):
            # - Cyber World
            config[Option.AMIGA_MODEL] = "A1200"
            config[Option.ACCURACY] = "0"
        config[Option.CHIP_MEMORY] = ""
        config[Option.SLOW_MEMORY] = ""
        config[Option.FAST_MEMORY] = "8192"


def read_whdload_args_from_info_data(data: bytes) -> List[str]:
    logging.debug("[WHDLOAD] Read WHDLoad args from info data")
    index = data.lower().find(b"slave=") - 1
    if not index:
        return []
    args = []
    parts = data[index:].split(b"\x00\x00\x00\x00")
    for part in parts[:]:
        if len(part) > 2:
            length = part[0]
            # logging.debug("%d", length)
            arg = part[1 : 1 + length - 1]
            if b"***" in arg:
                break
            args.append(arg.decode("ISO-8859-1"))
    return args


def read_whdload_args_from_info_stream(stream: BinaryIO) -> List[str]:
    return read_whdload_args_from_info_data(stream.read())


def strip_whdload_slave_prefix(whdload_args: List[str]) -> List[str]:
    result = []
    for i, arg in enumerate(whdload_args):
        arg = arg.split(";")[0]
        if i == 0 and arg.lower().startswith("slave="):
            arg = arg[6:]
        if not arg.startswith("("):
            result.append(arg)
    return result


def fix_whdload_args(args: List[str]) -> str:
    """Converts args list to string and corrects args for a few icons.

    The corrections remove binary garbage from the end of the arguments list
    (either there's a problem with the icons, or there's a bug in the
    .info "parser", the latter being the most likely.
    """
    args_str = " ".join(args)
    if args_str.startswith("KingsQuestEnhanced.Slave PRELOAD  /\x8c"):
        args_str = "KingsQuestEnhanced.Slave PRELOAD"
    elif args_str.startswith("PsychoKillerCDTV.Slave PRELOAD 3\x11"):
        args_str = "PsychoKillerCDTV.Slave PRELOAD"
    elif args_str.startswith("QuestForGlory.Slave PRELOAD \x00"):
        args_str = "QuestForGlory.Slave PRELOAD"
    elif args_str.startswith("Lorna.Slave PRELOAD \n"):
        args_str = "Lorna.Slave PRELOAD"
    elif args_str.startswith("BrainBlasters.Slave PRELOAD D\x88"):
        args_str = "BrainBlasters.Slave PRELOAD"
    elif args_str.startswith("MikroMortalTennis.slave PRELOAD fÿÝ"):
        args_str = "MikroMortalTennis.slave PRELOAD"
    elif args_str.startswith("BreakOutRevolutionAGA.slave PRELOAD \x0f"):
        args_str = "BreakOutRevolutionAGA.slave PRELOAD"
    elif args_str.startswith("KingsQuestEnhancedMT32.Slave PRELOAD /\x8c"):
        args_str = "KingsQuestEnhancedMT32.Slave PRELOAD"
    elif args_str.startswith("Turbo.Slave PRELOAD \x00"):
        args_str = "Turbo.Slave PRELOAD"
    elif args_str.startswith("WinterChallengeNTSC.slave PRELOAD NTSC î\x00"):
        args_str = "WinterChallengeNTSC.slave PRELOAD NTSC"
    elif args_str.startswith("OscarCD32NTSC.Slave PRELOAD NTSC Ì"):
        args_str = "OscarCD32NTSC.Slave PRELOAD NTSC"
    elif args_str.startswith("AlienBreed.Slave PRELOAD \x11"):
        args_str = "AlienBreed.Slave PRELOAD"
    elif args_str.startswith(
        "LittleComputerPeopleNTSC.Slave PRELOAD NTSC MAG\x00"
    ):
        # FIXME: Should MAG be included?
        args_str = "LittleComputerPeopleNTSC.Slave PRELOAD NTSC"
    return args_str.strip()


def calculate_whdload_args(archive_path: str) -> str:
    """
    This function, as it is currently written, only works if there
    is an .info with the same name as the .slave file. In theory, they
    could be different since the .info file contains a slave=... tool type.
    """
    archive = Archive(archive_path)
    slave_args = {}
    lower_to_name = {}
    for path in archive.list_files():
        path_lower = path.lower()
        if path_lower.rsplit("#/", 1)[1] == "s/startup-sequence":
            logging.info(
                "[WHDLOAD] Found Startup-Sequence, assuming "
                "non-WHDLoad archive"
            )
            return ""
        lower_to_name[path_lower] = path
    for path in lower_to_name.values():
        name = os.path.basename(path)
        name_lower = name.lower()
        if name_lower.endswith(".info"):
            try:
                args = read_whdload_args_from_info_stream(archive.open(path))
                args = strip_whdload_slave_prefix(args)
            except Exception as e:
                logging.warning(
                    "[WHDLOAD] WARNING: Error reading args: %s", repr(e)
                )
            else:
                if args:
                    archive_name = path.rsplit("#/", 1)[1]
                    logging.debug(
                        "[WHDLOAD] %s => %s".format(
                            archive_name, " ".join(args)
                        )
                    )
                    slave_args[archive_name] = args
    if len(slave_args) == 0:
        return ""
    if len(slave_args) > 1:
        logging.debug("[WHDLOAD] Multiple WHDLoad icons found")
        for icon, args in slave_args.items():
            if icon.lower() in primary_icons:
                logging.debug("[WHDLOAD] Choosing %s as primary icon", icon)
                return fix_whdload_args(args)
        raise Exception("Multiple slaves found")
    return fix_whdload_args(slave_args.popitem()[1])


def generate_config_for_archive(
    path: str, model_config: bool = True
) -> Dict[str, str]:
    logging.debug("[WHDLOAD] Generate config for archive %s", path)
    config = {}
    whdload_args = ""
    dummy, ext = os.path.splitext(path)
    if ext.lower() in Archive.extensions:
        try:
            whdload_args = calculate_whdload_args(path)
        except Exception:
            logging.exception("[WHDLOAD] Error while calculating WHDLoad args")
    config[Option.X_WHDLOAD_ARGS] = whdload_args
    if whdload_args and model_config:
        config[Option.AMIGA_MODEL] = "A1200"
        config[Option.FAST_MEMORY] = "8192"
        if should_disable_drive_click():
            config[Option.FLOPPY_DRIVE_VOLUME_EMPTY] = "0"
    return config


default_whdload_version = "18.5"

# noinspection SpellCheckingInspection
support_files = {
    "1d1c557f4a0f5ea88aeb96d68b09f41990340f70": "Devs/Kickstarts/kick33180.A500.RTB",
    "1ad1b55e7226bd5cd66def8370a69f19244da796": "Devs/Kickstarts/kick40068.A1200.RTB",
    "209c109855f94c935439b60950d049527d2f2484": "Devs/Kickstarts/kick34005.A500.RTB",
    "973b42dcaf8d6cb111484b3c4d3b719b15f6792d": "Devs/Kickstarts/kick40068.A4000.RTB",
    "ebf3a1f53be665bb39a636007fda3b3e640998ba": "C/uae-configuration",
    "51a37230cb45fc20fae422b8a60afd7cc8a63ed3": "C/OSEmu.400",
}
# noinspection SpellCheckingInspection
binaries = {
    "10.0": {"3096b2f41dfebf490aac015bdf0e91a80045c2c0": "C/WHDLoad"},
    "13.0": {"4bcb393e820d68b0520da9131e0d529018e303d1": "C/WHDLoad"},
    "16.0": {"883b9e37bc81fc081f78a3f278b732f97bdddf5c": "C/WHDLoad"},
    "16.1": {"250506c2444d9fb89b711b4fba5d70dd554e6f0e": "C/WHDLoad"},
    "16.2": {"a8bc2828c7da88f6236a8e82c763c71582f66cfd": "C/WHDLoad"},
    "16.3": {"5d636899fa9332b7dfccd49df3447238b5b71e49": "C/WHDLoad"},
    "16.4": {"1bb42fc83ee9237a6cfffdf15a3eb730504c9f65": "C/WHDLoad"},
    "16.5": {"8974e6c828ac18ff1cc29e56a31da0775ddeb0f0": "C/WHDLoad"},
    "16.6": {"b268bf7a05630d5b2bbf99616b32f282bac997bf": "C/WHDLoad"},
    "16.7": {"be94bc3d70d5980fac7fd04df996120e8220c1c0": "C/WHDLoad"},
    "16.8": {"a3286827c821386ac6e0bb519a7df807550d6a70": "C/WHDLoad"},
    "16.9": {"b4267a21918d6375e1bbdcaee0bc8b812e366802": "C/WHDLoad"},
    "17.0": {"0ec213a8c62beb3eb3b3509aaa44f21405929fce": "C/WHDLoad"},
    "17.1": {"1a907ca4539806b42ad5b6f7aeebacb3720e840d": "C/WHDLoad"},
    "17.2": {"d8f45f7808fb1ac356d88b8848660a6b96f04349": "C/WHDLoad"},
    "18.0": {"6f778e28673e9f931f81212ab03d9617a41cee40": "C/WHDLoad"},
    "18.1": {"fb4c64b0b5e682125e53eb2ace9bf0ccd3b8501f": "C/WHDLoad"},
    "18.2": {"0a9e7bfa1183420543e44c08410af1c5500fa704": "C/WHDLoad"},
    "18.3": {"b126b899d57f81a7646787776893416c20f43ec2": "C/WHDLoad"},
    "18.4": {"58efb174f54c67c7c48a7c1bff033c6d7a9884cb": "C/WHDLoad"},
    "18.5": {"d6b706bfbfe637bd98cd657114eea630b7d2dcc7": "C/WHDLoad"},
}
default_whdload_prefs = """
; wait for button pressed (slave must support this)
;ButtonWait

; disable cache-ability of Chip-Memory
;ChipNoCache

; path for core dump files
;CoreDumpPath=T:

; raw key code to quit with core dump (debug)
;DebugKey=$5b

; command to execute on WHDLoad startup
;ExecuteStartup=rx offline.rexx

; command to execute on WHDLoad exit
;ExecuteCleanup=rx online.rexx

; selects expert mode
;Expert

; raw key code to enter HrtMon/TK
;FreezeKey=$5d

; use MMU (for 68030)
;MMU

; ignore unwanted auto-vector interrupts
;NoAutoVec

; disable audio filter
;NoFilter

; do not flush memory
;NoFlushMem

; do not allocate memory reverse
;NoMemReverse

; disable the disk write cache
;NoWriteCache

;QuitKey=$45

; wait after reading from disk (1/50 seconds)
;ReadDelay=150

; raw key code to restart
;RestartKey=$5c

; command for Show Regs
;ShowRegs=SYS:Utilities/MuchMore W WL=80 WT=80 WW=582 WH=700

; time to display splash window (1/50 seconds)
;SplashDelay=0

; wait after saving something to disk (1/50 seconds)
;WriteDelay=150
"""
# noinspection SpellCheckingInspection
primary_icons = {
    _.lower()
    for _ in [
        "3DPool/3DPool.info",
        "AigleDOrFr/AigleDOrFr.info",
        "BartVsSpaceMutants2Disk/BartVsSpaceMutants2Disk.info",
        "BattleIsle/BattleIsle.info",
        "BattleIsle/Programm.info",
        "BattleIsle&DataDisks/BattleIsle.info",
        "BattleIsle&DataDisks/BattleIsle.info",
        "BattleIsle&DataDisks/Programm.info",
        "BlackViperCD32/BlackViperCD32.info",
        "Cadaver&CadaverThePayoff/CadaverThePayoff.info",
        "CastleOfDrBrain/CastleOfDrBrain.info",
        "ChaosStrikesBack/ChaosStrikesBack.info",
        "DGeneration/DGeneration.info",
        "DGenerationAGA/DGenerationAGA.info",
        "DGenerationCD32/DGenerationCD32.info",
        "EmeraldMine2/EmeraldMine2.info",
        "Entity/Entity.info",
        "EyeOfTheStorm/EyeOfTheStorm.info",
        "Fuzzball/Fuzzball.info",
        "Genesia/Genesia.info",
        "GenesiaFr/GenesiaFr.info",
        "HardDrivin2/HardDrivin2.info",
        "Historyline/Historyline.info",
        "Historyline/Historyline.info",
        "HistorylineDe/HistorylineDe.info",
        "HistorylineDe/HistorylineDe.info",
        "HistorylineFr/HistorylineFr.info",
        "HistorylineFr/HistorylineFr.info",
        "Jetstrike/Jetstrike.info",
        "JetstrikeAGA/JetstrikeAGA.info",
        "JimmyWillburneFr/JimmyWillburneFr.info",
        "Lemmings21MB/Lemmings21MB.info",
        "Lemmings2512KB/Lemmings2512KB.info",
        "MetalMasters/MetalMasters.info",
        "Might&Magic3/Might&Magic3.info",
        "Might&Magic3/Might&Magic3.info",
        "Might&Magic3De/Might&Magic3De.info",
        "MightAndMagic3/MightAndMagic3.info",
        "MightAndMagic3De/MightAndMagic3De.info",
        "MightAndMagic3De/MightAndMagic3.info",
        "PinballPrelude/PinballPrelude Past.info",
        "PinballPreludeAGA/PinballPreludeAGA Past.info",
        "PinballPreludeAGA/Past.info",
        "RaceDrivin/RaceDrivin.info",
        "SensibleSoccer/SensibleSoccer.info",
        "SensibleSoccer/SensibleSoccer.info",
        "SensibleSoccerEuro/SensibleSoccerEuro.info",
        "SensibleSoccerEuro/SensibleSoccerEuro.info",
        "SensibleSoccerIntEd/SensibleSoccerIntEd.info",
        "SensibleSoccerIntlEd/SensibleSoccerIntlEd.info",
        "SpaceHarrier&RFantasyZone/RFantasyZone.info",
        "SpaceHarrier&RetrnFntZone/ReturnToFantasyZone.info",
        "Superfrog/Superfrog.info",
        "SuperfrogCD32/SuperfrogCD32.info",
        "SuprStrtFtr2TrboAGA/SuprStrtFtr2TrboAGA.info",
        "SuprStrtFtr2TrboCD32/SuprStrtFtr2TrboCD32.info",
        "TubularWorldsECS/TubularWorldsECS.info",
        "TurboTraxArcane/TurboTraxArcane.info",
        "Virocop/Virocop2Meg.info",
        "WalkerCrunched/WalkerCrunched.info",
        "WalkerDecrunched/WalkerDecrunched.info",
    ]
}



from fsbc.util import unused
import fsui
from ...launcher_settings import LauncherSettings
from ...game_paths import GamePaths
from ..Constants import Constants
from ..ImageLoader import ImageLoader
from ..skin import Skin
from .BottomPanel import BottomPanel

BORDER = 20


class ScreenshotsPanel(BottomPanel):
    def __init__(self, parent):
        BottomPanel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.HorizontalLayout()
        self.image_loader = ImageLoader()

        def get_min_width():
            return 0

        # def get_min_height():
        #     return Constants.SCREEN_SIZE[1] + 2 * BORDER
        self.layout.get_min_width = get_min_width
        # self.layout.get_min_height = get_min_height
        self.layout.padding_left = BORDER // 2
        self.layout.padding_right = BORDER // 2
        self.layout.padding_top = BORDER + 2
        self.layout.padding_bottom = Skin.get_bottom_margin()

        self.default_image = fsui.Image("launcher:res/screenshot.png")
        # self.default_image.resize(Constants.SCREEN_SIZE)
        self.screenshot_overlay = fsui.Image(
            "launcher:res/screenshot_overlay.png"
        )

        self.images = [self.default_image for _ in range(6)]
        self.image_paths = ["" for _ in range(6)]
        self.requests = [None for _ in range(6)]

        self.x_offset = 0

        self.load_images()
        LauncherSettings.add_listener(self)

    def on_destroy(self):
        LauncherSettings.remove_listener(self)
        self.image_loader.stop()

    def set_min_screenshots(self, count):
        # w = SCREEN_SIZE[0] * count + BORDER * 2 + (BORDER + 1) * (count - 1)
        w = (
            Constants.SCREEN_SIZE[0] * count
            + BORDER * 2
            + BORDER * (count - 1)
        )
        self.set_min_width(w)

    def load_images(self):
        # t1 = time.time()
        handler = GamePaths.current()
        for i in range(6):
            path = handler.get_screenshot_path(i)
            if path == self.image_paths[i]:
                continue
            self.image_paths[i] = path

            def on_load(request):
                # print("on_load, request.image =", request.image)
                if request.path != self.image_paths[request.args["index"]]:
                    return
                if request.image:
                    self.images[request.args["index"]] = request.image
                else:
                    self.images[request.args["index"]] = self.default_image
                self.refresh()

            self.requests[i] = self.image_loader.load_image(
                path, size=Constants.SCREEN_SIZE, on_load=on_load, index=i
            )
            self.images[i] = self.default_image
            self.refresh()

            # image = handler.load_screenshot_preview(i)
            # if image:
            #     self.images[i] = image
            #     self.refresh()
            # else:
            #     self.images[i] = self.default_image
            #     self.refresh()

        # t2 = time.time()
        # print(t2 - t1)

    def on_setting(self, key, value):
        unused(value)
        if key == "config_name":
            self.x_offset = 0
            self.load_images()
        if key == "parent_uuid":
            self.x_offset = 0
            self.load_images()

    def on_left_down(self):
        # print("on_left_down")
        width = 22 + Constants.SCREEN_SIZE[0] * 6 + 22 * 5
        if self.x_offset == 0:
            self.x_offset = self.size()[0] - width
        else:
            self.x_offset = 0
        self.refresh()

    def on_paint(self):
        dc = self.create_dc()
        self.draw_background(dc)
        size = self.size()

        y = 2 + 20
        x = 10 + self.x_offset
        for i in range(6):
            if x >= size[0] - 12:
                break
            image = self.images[i]
            # dc.draw_image(image, x + 1, y + 1, Constants.SCREEN_SIZE[0],
            #         Constants.SCREEN_SIZE[1])
            dc.draw_image(image, x + 1, y + 1)
            dc.draw_image(self.screenshot_overlay, x - 10, y - 10)
            x = x + Constants.SCREEN_SIZE[0] + 22



import traceback
import fsui


class ServerWindow(fsui.Window):
    def __init__(self, parent, server):
        fsui.Window.__init__(self, parent, "FS-UAE Net Play Server")
        self.layout = fsui.VerticalLayout()
        self.layout.padding_top = 50
        self.layout.padding_bottom = 50
        self.layout.padding_left = 50
        self.layout.padding_right = 50
        self.label = fsui.Label(self, "Close window to stop server")
        self.layout.add(self.label)
        self.server = server
        self.set_size(self.layout.get_min_size())

    def on_close(self):
        print("ServerWindow.on_close")
        try:
            self.server.kill()
        except Exception:
            traceback.print_exc()



from fsui.qt import Qt, QSize, Signal, QFont
from fsui.qt import QListView, QStandardItemModel, QStandardItem
from .widget_mixin import WidgetMixin


class ListView(QListView, WidgetMixin):

    item_selected = Signal(int)
    item_activated = Signal(int)

    def __init__(self, parent, border=True):
        # self = QListView(parent.get_container())
        QListView.__init__(self, parent.get_container())
        # Widget.__init__(self, parent)
        # self.setAutoFillBackground(True)
        self.init_widget(parent)
        self.viewport().installEventFilter(self.get_window())
        self.verticalScrollBar().installEventFilter(self.get_window())
        if not border:
            self.setFrameStyle(0)

        # self.setSelectionModel()
        self._model = QStandardItemModel(self)
        # self.setModel(self._model)
        self.setModel(self._model)
        # self.itemSelectionChanged.connect(self._on_selection_changed)
        selection_model = self.selectionModel()
        print("QListView selectionModel", selection_model)
        selection_model.selectionChanged.connect(self.__selection_changed)
        self.setEditTriggers(QListView.NoEditTriggers)
        self.doubleClicked.connect(self.__double_clicked)
        # self.returnPressed.connect(self.__double_clicked)
        # self.activated.connect(self.__double_clicked)
        self._row_height = 26

    def set_row_height(self, height):
        self._row_height = height

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return:
            self.__double_clicked()
        else:
            super().keyPressEvent(event)

    def __double_clicked(self):
        index = self.get_index()
        if index is not None:
            self.on_activate_item(index)
            self.item_activated.emit(index)

    def on_activate_item(self, index):
        pass

    def __selection_changed(self):
        index = self.get_index()
        self.on_select_item(index)
        self.item_selected.emit(index)

    def set_default_icon(self, image):
        pass

    def set_items(self, items):
        self._model.clear()
        for item in items:
            if isinstance(item, str):
                self.add_item(item)
            elif isinstance(item, dict):
                label = item["label"]
                icon = item.get("icon", None)
                bold = item.get("bold", False)
                self.add_item(label, icon, bold)
            else:
                label, icon = item
                self.add_item(label, icon)

    def clear(self):
        self._model.clear()

    def add_item(self, label, icon=None, bold=False):
        item = QStandardItem(label)
        if icon:
            try:
                item.setIcon(icon.qicon(16))
            except TypeError:
                item.setIcon(icon.qicon)
        item.setSizeHint(QSize(-1, self._row_height))
        if bold:
            font = self.font()
            font.setWeight(QFont.Bold)
            item.setFont(font)
        self._model.appendRow(item)

    def get_item(self, index):
        return self._model.item(index).text()

    def get_item_count(self):
        return self._model.rowCount()

    def get_index(self):
        indices = self.selectionModel().selectedIndexes()
        if len(indices) == 0:
            return None
        return indices[0].row()

    def set_index(self, index):
        if index is None:
            index = -1
        # print(self.rootIndex)
        # idx = QModelIndex.createIndex(index)
        idx = self._model.index(index, 0)
        self.scrollTo(idx)
        self.setCurrentIndex(idx)

    def select_item(self, index):
        self.set_index(index)

    def on_select_item(self, index):
        print("calling item_selected.emit")
        self.item_selected.emit(index)



#!/usr/bin/env python3

import os
import sys
import hashlib
from io import BytesIO

# noinspection PyUnresolvedReferences
from typing import List, Dict


B_SIZE = 512
B_COUNT = 880 * 2
T_HEADER = 2
T_DATA = 8
T_LIST = 16
ST_ROOT = 1
ST_USERDIR = 2
ST_FILE = 0xFFFFFFFF - 3 + 1


def char(block: bytes, pos: int) -> str:
    return block[pos : pos + 1].decode("ISO-8859-1")


def ubyte(block: bytes, pos: int) -> int:
    return block[pos]


def ulong(block: bytes, pos: int) -> int:
    a = block[pos]
    b = block[pos + 1]
    c = block[pos + 2]
    d = block[pos + 3]
    return (a << 24) | (b << 16) | (c << 8) | d


def checksum_block(block: bytes) -> int:
    # clear checksum field
    data = block[:20] + b"\0\0\0\0" + block[24:]
    sum = 0
    for pos in range(0, B_SIZE, 4):
        sum += ulong(data, pos)
        sum &= 0xFFFFFFFF
    sum = -sum
    sum &= 0xFFFFFFFF
    return sum


def verify_block(data: bytes) -> bool:
    sum = 0
    for pos in range(0, B_SIZE, 4):
        sum += ulong(data, pos)
        sum &= 0xFFFFFFFF
    return sum == 0


class Block(object):
    def __init__(self, data: bytes) -> None:
        self.data = data

    def char(self, pos: int) -> str:
        return char(self.data, pos)

    def string(self, pos: int, len: int) -> str:
        return self.data[pos : pos + len].decode("ISO-8859-1")

    def ubyte(self, pos: int) -> int:
        return ubyte(self.data, pos)

    def ulong(self, pos: int) -> int:
        return ulong(self.data, pos)


class FileInfo(object):
    def __init__(self):
        self.block_list = []
        self.header_block = -1
        self.name = ""
        self.path = ""
        self.comment = None
        self.time = ""
        self.mode = 0
        self.size = 0


class ADFFile(object):

    FFS_FLAG = 1
    INTL_ONLY_FLAG = 2
    DIRC_AND_INTL_FLAG = 4

    def __init__(self, stream_or_data_or_file):
        if hasattr(stream_or_data_or_file, "read"):
            data = stream_or_data_or_file.read()
        elif len(stream_or_data_or_file) == B_SIZE * B_COUNT:
            data = stream_or_data_or_file
        else:
            with open(stream_or_data_or_file, "rb") as f:
                data = f.read()
        assert isinstance(data, bytes)
        # print(len(data))
        assert len(data) == B_SIZE * B_COUNT
        self.blocks = []  # type: List[Block]
        for i in range(0, B_SIZE * B_COUNT, B_SIZE):
            self.blocks.append(Block(data[i : i + B_SIZE]))
        assert len(self.blocks) == B_COUNT
        self.block_usage = [["block"] for _ in range(B_COUNT)]
        self.dos = False
        self.ofs = False
        self.ffs = False
        self.warnings = []  # type: List[str]
        self.file_map = {}  # type: Dict[str, FileInfo]
        self.root_block_number = 880
        self.bitmap_pages = []
        self._parse()

    def root_block(self) -> Block:
        return self.blocks[self.root_block_number]

    def _parse(self) -> None:
        b = self.blocks[0]
        if b.char(0) == "D" and b.char(1) == "O" and b.char(2) == "S":
            self.dos = True
        else:
            return
        flags = b.ubyte(3)
        self.ffs = flags & self.FFS_FLAG
        self.ofs = not self.ffs

        self.root_block_number = b.ulong(8)
        self.bitmap_pages = []
        if self.root_block_number != 880:
            self.warnings.append(
                "Root block is at position {0}, "
                "not 880".format(self.root_block_number)
            )
            self.root_block_number = 880
            self.warnings.append("Trying 880 anyway...")

        self._parse_root_block()
        self.block_usage[self.root_block_number].append("root block")

        for i, usage in enumerate(self.block_usage):
            if "used" in usage:
                if usage == ["block", "used"]:
                    self.warnings.append(
                        "block {0} marked as used but no actual usage".format(
                            i
                        )
                    )
            if "free" in usage:
                if usage != ["block", "free"]:
                    self.warnings.append(
                        "block {0} marked as used but used for {1}".format(
                            i, repr(usage)
                        )
                    )

    def _parse_root_block(self) -> None:
        b = self.root_block()
        type = b.ulong(0)
        secondary_type = b.ulong(B_SIZE - 4)
        if type != T_HEADER:
            print(type, T_HEADER)
            self.warnings.append("root block does not have T_HEADER type")
        if secondary_type != ST_ROOT:
            self.warnings.append(
                "root block does not have ST_ROOT secondary " "type"
            )

        days = b.ulong(B_SIZE - 92)
        mins = b.ulong(B_SIZE - 88)
        ticks = b.ulong(B_SIZE - 84)
        self.root_mtime = "{0:05d}:{1:04d}:{2:03d}".format(days, mins, ticks)

        days = b.ulong(B_SIZE - 40)
        mins = b.ulong(B_SIZE - 36)
        ticks = b.ulong(B_SIZE - 32)
        self.disk_mtime = "{0:05d}:{1:04d}:{2:03d}".format(days, mins, ticks)

        days = b.ulong(B_SIZE - 28)
        mins = b.ulong(B_SIZE - 24)
        ticks = b.ulong(B_SIZE - 20)
        self.disk_ctime = "{0:05d}:{1:04d}:{2:03d}".format(days, mins, ticks)

        bm_flag = b.ulong(B_SIZE - 200)
        if bm_flag != 0xFFFFFFFF:
            self.warnings.append(
                "bm_flag != 0xffffffff ({0:8x})".format(bm_flag)
            )
            if bm_flag == 0:
                self.warnings.append("bm_flag is ZERO")
        for i in range(25):
            self.bitmap_pages.append(b.ulong(B_SIZE - 196 + 4 * i))
        self._parse_used_blocks()

        name_len = b.ubyte(B_SIZE - 80)
        self.volume_name = b.string(B_SIZE - 79, name_len)
        # self.volume_name = self.volume_name.decode("ISO-8859-1")
        self._parse_directory_content("", self.root_block_number)
        if not verify_block(b.data):
            self.warnings.append("bad root checksum")

    def _parse_used_blocks(self) -> None:
        for i in range(len(self.bitmap_pages)):
            block_number = self.bitmap_pages[i]
            if block_number:
                if block_number >= B_COUNT:
                    self.warnings.append(
                        "invalid block number {0} in bitmap pages".format(
                            block_number
                        )
                    )
                    continue
                self.block_usage[block_number].append("bitmap block")

            if i == 0:
                if not block_number:
                    self.warnings.append("no bitmap block for disk")
                    continue
            else:
                if block_number:
                    self.warnings.append(
                        "unexpected additional bitmap block..."
                    )
                continue

            b = self.blocks[block_number]
            if not verify_block(b.data):
                self.warnings.append(
                    "bitmap block checksum is invalid (" "ignoring this block)"
                )
                continue
            for bit_index in range(1758):
                long_index = bit_index // 32
                bit = bit_index % 32
                ul = b.ulong(4 + long_index * 4)
                free = ul & (1 << bit)
                if free:
                    self.block_usage[2 + bit_index].append("free")
                else:
                    self.block_usage[2 + bit_index].append("used")

    def _parse_directory_content(self, path: str, block_number: int) -> None:
        d_b = self.blocks[block_number]
        d_secondary_type = d_b.ulong(B_SIZE - 4)
        if d_secondary_type == ST_ROOT:
            ht_size = d_b.ulong(12)
            assert ht_size == (B_SIZE // 4) - 56

        for i in range((B_SIZE // 4) - 56):
            entry = d_b.ulong(24 + i * 4)
            while entry:
                # print(entry)
                if entry >= B_COUNT:
                    self.warnings.append(
                        "directory entry refers to an invalid block "
                        "number {0}".format(entry)
                    )
                    entry = 0
                    continue
                b = self.blocks[entry]
                type = b.ulong(0)
                secondary_type = b.ulong(B_SIZE - 4)
                parent = b.ulong(B_SIZE - 12)
                assert type == T_HEADER
                assert secondary_type in [ST_USERDIR, ST_FILE]
                assert parent == block_number

                if secondary_type == ST_USERDIR:
                    self._parse_directory(path, entry)
                elif secondary_type == ST_FILE:
                    self._parse_file(path, entry)
                else:
                    raise Exception("neither ST_USERDIR or ST_FILE")

                entry = b.ulong(B_SIZE - 16)

    def _parse_file(self, path: str, block_number: int) -> None:
        b = self.blocks[block_number]
        header_key = b.ulong(4)
        assert header_key == block_number

        file_info = FileInfo()
        file_info.header_block = block_number

        name_len = b.ubyte(B_SIZE - 80)
        file_info.name = b.string(B_SIZE - 79, name_len)
        file_info.path = path + file_info.name
        # print(file_info.path)
        file_key = file_info.path.lower()
        if file_key in self.file_map:
            self.warnings.append(
                "duplicate entries for file name " "{0}".format(file_key)
            )
        self.file_map[file_key] = file_info

        self.block_usage[block_number].append(
            "header block for file " + file_info.path
        )
        if not verify_block(b.data):
            self.warnings.append(
                "bad checksum for header for " + file_info.path
            )

        file_info.mode = b.ulong(B_SIZE - 192)
        file_info.size = b.ulong(B_SIZE - 188)
        comment_len = b.ubyte(B_SIZE - 184)
        file_info.comment = b.string(B_SIZE - 183, comment_len)

        days = b.ulong(B_SIZE - 92)
        mins = b.ulong(B_SIZE - 88)
        ticks = b.ulong(B_SIZE - 84)
        file_info.time = "{0:05d}:{1:04d}:{2:03d}".format(days, mins, ticks)

        first_data = b.ulong(16)
        next_data = first_data
        file_blocks_1 = []  # type: List[int]
        ofs_accum_size = 0
        k = 0

        while next_data and self.ofs:
            file_blocks_1.append(next_data)
            data_b = self.blocks[next_data]
            data_type = data_b.ulong(0)
            assert data_type == T_DATA
            header_key = data_b.ulong(4)
            if header_key != block_number:
                self.warnings.append("header_key != block_number")

            seq_num = data_b.ulong(8)
            if seq_num != len(file_blocks_1):
                self.warnings.append(
                    "block {0:04d} expected seq_num {1} (found {2}) for "
                    "file {3}".format(
                        next_data, len(file_blocks_1), seq_num, file_info.path
                    )
                )
            data_size = data_b.ulong(12)
            ofs_accum_size += data_size

            if not verify_block(data_b.data):
                self.warnings.append(
                    "block {1:04d} - bad checksum for data block #{0} for "
                    "file {2}".format(k + 1, next_data, file_info.path)
                )
            # self.block_usage[next_data].append(
            #     "block {2:04d} - data block #{0} for file {1}".format(
            #         k + 1, file_info.path, next_data))

            next_data = data_b.ulong(16)
            k += 1

        file_blocks_2 = []  # type: List[int]
        extension = block_number
        k = 0
        while extension:
            ext_b = self.blocks[extension]
            ext_type = ext_b.ulong(0)
            ext_secondary_type = ext_b.ulong(B_SIZE - 4)
            if extension != block_number:
                assert ext_type == T_LIST
            assert ext_secondary_type == ST_FILE
            high_seq = ext_b.ulong(8)
            for i in range(high_seq):
                file_blocks_2.append(ext_b.ulong(B_SIZE - 204 - i * 4))

            parent = ext_b.ulong(B_SIZE - 12)
            if k > 0 and parent != block_number:
                self.warnings.append(
                    "block {3:04d} - expected parent {0} (found {1} for OFS "
                    "extension block #{2} for file {4}".format(
                        block_number, parent, k, extension, file_info.path
                    )
                )

            if not verify_block(ext_b.data):
                self.warnings.append(
                    "block {2:04d} - bad checksum for extension block #{0} "
                    "for file {1}".format(k, file_info.path, extension)
                )
            extension = ext_b.ulong(B_SIZE - 8)
            if extension:
                self.block_usage[extension].append(
                    "file extension block #{0} for file {1}".format(
                        k, file_info.path
                    )
                )
            k += 1

        if self.ofs:
            if file_blocks_1 != file_blocks_2:
                self.warnings.append(
                    "block list mismatch (extension vs OFS headers) for "
                    "file {0}".format(file_info.path)
                )
            if len(file_blocks_1) != len(file_blocks_2):
                self.warnings.append(
                    "block list length mismatch (extension {0} vs OFS "
                    "headers {1}) for file {2}".format(
                        len(file_blocks_2), len(file_blocks_1), file_info.path
                    )
                )
            if ofs_accum_size != file_info.size:
                self.warnings.append(
                    "file size mismatch (file header {0} vs OFS data block "
                    "headers {1}) for file {2}".format(
                        file_info.size, ofs_accum_size, file_info.path
                    )
                )

        file_info.block_list = file_blocks_2

        for i, bn in enumerate(file_blocks_2):
            self.block_usage[bn].append(
                "data block #{0} for file {1}".format(
                    i + 1, file_info.path, bn
                )
            )

    def _parse_directory(self, path: str, block_number: int) -> None:
        b = self.blocks[block_number]
        header_key = b.ulong(4)
        assert header_key == block_number

        file_info = FileInfo()
        file_info.header_block = block_number

        name_len = b.ubyte(B_SIZE - 80)
        file_info.name = b.string(B_SIZE - 79, name_len)
        file_info.path = path + file_info.name + "/"
        # print(file_info.path)
        file_key = file_info.path.lower()
        if file_key[:-1] in self.file_map:
            self.warnings.append(
                "Duplicate entries for file name " "{0}".format(file_key)
            )
        self.file_map[file_key] = file_info

        self.block_usage[block_number].append(
            "header block for file " + file_info.path
        )
        if not verify_block(b.data):
            self.warnings.append(
                "bad checksum for header for " + file_info.path
            )

        file_info.mode = b.ulong(B_SIZE - 192)
        file_info.size = 0
        comment_len = b.ubyte(B_SIZE - 184)
        file_info.comment = b.string(B_SIZE - 183, comment_len)

        days = b.ulong(B_SIZE - 92)
        mins = b.ulong(B_SIZE - 88)
        ticks = b.ulong(B_SIZE - 84)
        file_info.time = "{0:05d}:{1:04d}:{2:03d}".format(days, mins, ticks)

        self._parse_directory_content(file_info.path, block_number)

    def namelist(self) -> List[str]:
        names = []  # type: List[str]
        keys = sorted(self.file_map.keys())
        for key in keys:
            names.append(self.file_map[key].path)
        return names

    def getinfo(self, name: str) -> FileInfo:
        name = name.lower()
        return self.file_map[name]

    def open(self, name: str, mode: str = "r") -> BytesIO:
        assert mode == "r"
        return BytesIO(self.read(name))

    def read(self, name: str) -> bytes:
        name = name.lower()
        # print(repr(self.file_map))
        # print(repr(name))
        file_info = self.file_map[name]
        bytes_left = file_info.size
        data = []  # type: List[bytes]
        for block_number in file_info.block_list:
            if self.ffs:
                start_index = 0
            else:
                start_index = 24
            max_bsize = B_SIZE - start_index
            read_size = min(bytes_left, max_bsize)
            block_data = self.blocks[block_number].data[
                start_index : start_index + read_size
            ]
            bytes_left -= len(block_data)
            data.append(block_data)
        assert bytes_left == 0
        return b"".join(data)


def main():
    adf = ADFFile(sys.argv[1])
    print("")
    print(os.path.basename(sys.argv[1]))
    print("Volume:", adf.volume_name, "(OFS)" if adf.ofs else "(FFS)")
    print("")
    for name in adf.namelist():
        print(name)
        if not name.endswith("/"):
            data = adf.open(name).read()
            print("   ", hashlib.sha1(data).hexdigest(), "", len(data))
            # tracks = set()
            # for block in adf.file_map[name.lower()].block_list:
            #     track = block // 11
            #     cylinder = track // 2
            #     side = track % 2
            #     tracks.add((cylinder, side))
            # tracks = sorted(tracks)
            # print(tracks)
    print("")
    for warning in adf.warnings:
        print("WARNING:", warning)
    pass
    # for i, usage in enumerate(adf.block_usage):
    #     print(i, usage)
    for name in sys.argv[2:]:
        data = adf.open(name).read()
        with open(name + ".dump", "wb") as f:
            f.write(data)
            print("wrote " + name + ".dump")


if __name__ == "__main__":
    main()



import os
from binascii import unhexlify
import sqlite3

# FIXME: remove dependency on launcher, have the Launcher tell this class
# the path instead
from fsgs.FSGSDirectories import FSGSDirectories
from .BaseDatabase import BaseDatabase

VERSION = 1


class LockerDatabase(BaseDatabase):
    def __init__(self, sentinel):
        BaseDatabase.__init__(self, sentinel)

    @classmethod
    def get_path(cls):
        return os.path.join(FSGSDirectories.databases_dir(), "Locker.sqlite")

    @classmethod
    def get_version(cls):
        return VERSION

    @classmethod
    def instance(cls, new=False):
        if new or not hasattr(cls.thread_local, "locker_database"):
            cls.thread_local.locker_database = cls(cls.SENTINEL)
        return cls.thread_local.locker_database

    def check_sha1(self, sha1):
        cursor = self.internal_cursor()
        # FIXME: Replace with EXISTS or something
        cursor.execute(
            "SELECT count(*) FROM file WHERE sha1 = ?",
            (sqlite3.Binary(unhexlify(sha1.encode("ASCII"))),),
        )
        return cursor.fetchone()[0]

    def get_sync_version(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT sync_version FROM metadata")
        row = cursor.fetchone()
        return row[0]

    def set_sync_version(self, sync_version):
        cursor = self.internal_cursor()
        cursor.execute("UPDATE metadata set sync_version = ?", (sync_version,))

    def clear(self):
        if not self.connection:
            self.init()
        cursor = self.internal_cursor()
        cursor.execute("DELETE FROM file")

    def add_sha1_binary(self, sha1):
        cursor = self.internal_cursor()
        cursor.execute(
            "INSERT INTO file (sha1) VALUES (?)", (sqlite3.Binary(sha1),)
        )

    def update_database_to_version_1(self):
        cursor = self.internal_cursor()
        try:
            cursor.execute("SELECT count(*) FROM file")
        except sqlite3.OperationalError:
            cursor.execute(
                """CREATE TABLE file (
                    sha1 BLOB PRIMARY KEY
                    )"""
            )

        cursor.execute("ALTER TABLE metadata ADD COLUMN sync_version TEXT")



from fsbc.resources import Resources
from fsui.qt import Qt, QImage, QIcon, QPixmap


class Image(object):

    NEAREST = 0

    def __init__(self, name="", object=None):
        if object:
            self.qimage = object
        else:
            self.qimage = QImage()

            if hasattr(name, "read"):
                self.qimage.loadFromData(name.read())
            elif name.startswith("pkg://"):
                parts = name.split("/", 3)
                stream = Resources(parts[2]).stream(parts[3])
                self.qimage.loadFromData(stream.read())
            else:
                index = name.find(":")
                if index > 1:
                    package, file_ = name.split(":", 1)
                    stream = Resources(package).stream(file_)
                    self.qimage.loadFromData(stream.read())
                else:
                    print("loading image from", name)
                    self.qimage.load(name)
            # self._bitmap = None

    @property
    def size(self):
        return self.qimage.width(), self.qimage.height()

    def width(self):
        return self.qimage.width()

    def height(self):
        return self.qimage.height()

    @property
    def qpixmap(self):
        return QPixmap(self.qimage)

    @property
    def qicon(self):
        return QIcon(QPixmap(self.qimage))

    # @property
    # def bitmap(self):
    #     if self._bitmap is None:
    #         self._bitmap = wx.BitmapFromImage(self.qimage)
    #     return self._bitmap

    def grey_scale(self):
        # return Image(object=self.qimage.convertToFormat(
        #     QImage.Format_ARGB32, Qt.AutoOnly))
        copy = self.qimage.convertToFormat(QImage.Format_ARGB32, Qt.AutoColor)
        # copy = self.qimage.copy(0, 0, *self.size)

        # WARNING: this is presumably a bit slow...
        for y in range(self.size[1]):
            for x in range(self.size[0]):
                p = copy.pixel(x, y)

                # RGBA
                # r = (p & 0xff000000) >> 24
                # g = (p & 0x00ff0000) >> 16
                # b = (p & 0x0000ff00) >> 8
                # a = p & 0x000000ff
                # # v = (r + g + b) // 3
                # v = int(r * 0.299 + g * 0.587 + b * 0.114)
                # p = v << 24 | v << 16 | v << 8 | a

                # ARGB
                a = (p & 0xFF000000) >> 24
                r = (p & 0x00FF0000) >> 16
                g = (p & 0x0000FF00) >> 8
                b = p & 0x000000FF
                # v = (r + g + b) // 3
                v = int(r * 0.299 + g * 0.587 + b * 0.114)
                p = a << 24 | v << 16 | v << 8 | v

                copy.setPixel(x, y, p)
        return Image(object=copy)

    def resize(self, size, filter=1):
        if size == self.size:
            return
        if filter:
            q = Qt.SmoothTransformation
        else:
            q = Qt.FastTransformation
        self.qimage = self.qimage.scaled(
            size[0], size[1], Qt.IgnoreAspectRatio, q
        )
        # self._bitmap = None



# FSGS - Common functionality for Fengestad Game System.
# Copyright (C) 2013-2019  Frode Solheim <frode@solheim.dev>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import os

from fsgs.option import Option
from fsgs.platforms.loader import SimpleLoader

ZXS_MODEL_48K = "spectrum"
ZXS_MODEL_128 = "spectrum128"
ZXS_MODEL_PLUS3 = "spectrum+3"


class SpectrumLoader(SimpleLoader):
    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        if len(file_list) == 0:
            self.config["x_variant_error"] = "Variant has empty file list"
        elif len(file_list) > 1:
            self.config["x_variant_error"] = "Unsupported multi-file variant"

        name = file_list[0]["name"]
        sha1 = file_list[0]["sha1"]
        _, ext = os.path.splitext(name)
        ext = ext.lower()
        if ext in [".z80"]:
            key = Option.SNAPSHOT_FILE
        elif ext in [".tap", ".tzx"]:
            key = Option.TAPE_DRIVE_0
        elif ext in [".dsk", ".trd"]:
            key = Option.FLOPPY_DRIVE_0
        elif ext in [".rom"]:
            key = Option.CARTRIDGE_SLOT
        else:
            return
        self.config[key] = "sha1://{0}/{1}".format(sha1, name)

    def load_extra(self, values):
        self.config[Option.ZXS_MODEL] = values["model"]
        if not self.config[Option.ZXS_MODEL]:
            self.config[Option.ZXS_MODEL] = ZXS_MODEL_48K
        self.config["model"] = ""



from fsui.qt import QComboBox, Qt
from fsui.qt.signal import Signal, SignalWrapper
from .widget_mixin import WidgetMixin


class Choice(QComboBox, WidgetMixin):

    changed_signal = Signal()
    item_selected = Signal(int)
    ITEM_SEPARATOR = "---"

    def __init__(self, parent, items=None, cursor_keys=True):
        if items is None:
            items = []
        QComboBox.__init__(self, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        self.inhibit_change_event = False
        self.cursor_keys = cursor_keys

        for i, item in enumerate(items):
            self.insertItem(i, item)

        if len(items) > 0:
            self.set_index(0)
        self.currentIndexChanged.connect(self.__current_index_changed)

        self.changed = SignalWrapper(self, "changed")
        # self.changed.inhibit = self.inhibit_signal

    def keyPressEvent(self, event):
        if not self.cursor_keys:
            print("cursor keys is false", event.key(), Qt.Key_Up)
            if event.key() == Qt.Key_Up or event.key() == Qt.Key_Down:
                print("ignoring")
                return
        super().keyPressEvent(event)

    # @contextmanager
    # def inhibit_signal(self, name):
    #     attr = "_inhibit_" + name
    #     old = getattr(self, attr, False)
    #     print("setattr", self, attr, True)
    #     setattr(self, attr, True)
    #     yield
    #     print("setattr", self, attr, old)
    #     setattr(self, attr, old)

    def add_item(self, label, icon=None):
        # item = QStandardItem(label)
        # if icon:
        #     item.setIcon(icon.qicon)
        # item.setSizeHint(QSize(-1, 24))
        if label == self.ITEM_SEPARATOR:
            self.insertSeparator(self.count())
        elif icon is not None:
            self.addItem(icon.qicon, label)
        else:
            self.addItem(label)
        return self.count() - 1

    def remove_item(self, index):
        self.removeItem(index)

    def __current_index_changed(self):
        # print("__current_index_changed", self.currentIndex(),
        #       "inhibit", self.inhibit_change_event)
        if not self.inhibit_change_event:
            # print("Choice.__current_index_changed")
            # if not getattr(self, "_inhibit_changed", False):
            if not self.changed.inhibit:
                if not getattr(self, "_inhibit_item_selected", False):
                    index = self.currentIndex()
                    self.item_selected.emit(index)
                self.changed.emit()
                self.on_changed()

    def get_index(self):
        return self.currentIndex()

    def set_index(self, index, signal=True):
        try:
            if not signal:
                self.inhibit_change_event = True
            self.setCurrentIndex(-1 if index is None else index)
        finally:
            if not signal:
                self.inhibit_change_event = False

    def set_item_text(self, index, text):
        self.setItemText(index, text)

    def on_changed(self):
        pass

    def __len__(self):
        return self.count()


class ItemChoice(Choice):
    def __init__(self, parent):
        Choice.__init__(self, parent)

    def update(self):
        # for i, item in enumerate(self.items):
        #     self.add_item(item["name"], icon=self.get_item_icon(i))
        old = self.inhibit_change_event
        old_index = self.get_index()
        self.inhibit_change_event = True
        self.clear()
        for i in range(self.get_item_count()):
            text = self.get_item_text(i)
            icon = self.get_item_icon(i)
            self.add_item(text, icon)
        if old_index < self.get_item_count():
            self.set_index(old_index)
        self.inhibit_change_event = old

    def select_item(self, index, signal=True):
        print("select_item", index)
        if index is None:
            self.set_index(-1, signal=signal)
        else:
            self.set_index(index, signal=signal)

    def on_changed(self):
        self.on_select_item(self.get_index())

    def on_select_item(self, index):
        pass



import filecmp
import os

import shutil

from fsgs.drivers.messdriver import MessDriver
from fsgs.knownfiles import KnownFile
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader
from fsgs.plugins.pluginmanager import PluginManager

# noinspection SpellCheckingInspection
A5200_ROM = KnownFile(
    "6ad7a1e8c9fad486fbec9498cb48bf5bc3adc530", "A5200", "5200.rom"
)
# noinspection SpellCheckingInspection
A5200_ROM_ALT = KnownFile(
    "1d2a3f00109d75d2d79fecb565775eb95b7d04d5", "A5200", "5200a.rom"
)
A5200_CONTROLLER = {
    "type": "controller",
    "description": "Controller",
    "mapping_name": "atari5200",
}


class Atari5200PlatformHandler(Platform):
    PLATFORM_ID = "A5200"
    PLATFORM_NAME = "Atari 5200"

    def driver(self, fsgs):
        return Atari5200MessDriver(fsgs)

    def loader(self, fsgs):
        return Atari5200Loader(fsgs)


class Atari5200Loader(SimpleLoader):
    pass


class Atari5200MessDriver(MessDriver):
    PORTS = [
        {"description": "Input Port 1", "types": [A5200_CONTROLLER]},
        {"description": "Input Port 2", "types": [A5200_CONTROLLER]},
    ]

    def prepare(self):
        super().prepare()
        self.install_mame_hash_file("a5200.hsi")
        self.install_mame_hash_file("a5200.xml")

    def get_game_refresh_rate(self):
        # There were no PAL version of Atari 5200
        return 59.94

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, port):
        return {
            "START": 'type="P#_START"',
            "PAUSE": 'tag="keypad_2" type="KEYPAD" mask="1" defvalue="0"',
            "RESET": 'tag="keypad_1" type="KEYPAD" mask="1" defvalue="0"',
            "RIGHT": ('type="P#_AD_STICK_X"', "increment"),
            "LEFT": ('type="P#_AD_STICK_X"', "decrement"),
            "UP": ('type="P#_AD_STICK_Y"', "decrement"),
            "DOWN": ('type="P#_AD_STICK_Y"', "increment"),
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
            "PAD#": 'tag="keypad_0" type="KEYPAD" mask="2" defvalue="0"',
            "PAD0": 'tag="keypad_0" type="KEYPAD" mask="4" defvalue="0"',
            "PAD*": 'tag="keypad_0" type="KEYPAD" mask="8" defvalue="0"',
            "PAD9": 'tag="keypad_1" type="KEYPAD" mask="2" defvalue="0"',
            "PAD8": 'tag="keypad_1" type="KEYPAD" mask="4" defvalue="0"',
            "PAD7": 'tag="keypad_1" type="KEYPAD" mask="8" defvalue="0"',
            "PAD6": 'tag="keypad_2" type="KEYPAD" mask="2" defvalue="0"',
            "PAD5": 'tag="keypad_2" type="KEYPAD" mask="4" defvalue="0"',
            "PAD4": 'tag="keypad_2" type="KEYPAD" mask="8" defvalue="0"',
            "PAD3": 'tag="keypad_3" type="KEYPAD" mask="2" defvalue="0"',
            "PAD2": 'tag="keypad_3" type="KEYPAD" mask="4" defvalue="0"',
            "PAD1": 'tag="keypad_3" type="KEYPAD" mask="8" defvalue="0"',
        }

    def mess_romset(self):
        return (
            "a5200",
            {A5200_ROM.sha1: "5200.rom", A5200_ROM_ALT.sha1: "5200a.rom"},
        )



import gettext as gettext_module
import locale
import os
import sys
import traceback

import fsboot
from fsbc.settings import Settings

_initialized = False


def initialize_locale(language=None):
    global translations, _initialized
    _initialized = True

    if language is None:
        language = Settings.instance()["language"]

    print("[I18N] Initialize_locale language =", language)
    loc = language
    if not loc:
        try:
            loc, _charset = locale.getdefaultlocale()
            print("[I18N] Locale is", loc)
        except:
            print("[I18N] Exception while checking locale")
            loc = ""
        if not loc:
            loc = ""

    if not loc and sys.platform == "darwin":
        try:
            # noinspection PyUnresolvedReferences
            import CoreFoundation

            c_loc = CoreFoundation.CFLocaleCopyCurrent()
            loc = CoreFoundation.CFLocaleGetIdentifier(c_loc)
        except Exception:
            traceback.print_exc()
        print("[I18N] OS X locale", loc)

    dirs = [
        os.path.join(fsboot.executable_dir(), "share"),
        os.path.join(fsboot.executable_dir(), "..", "share"),
    ]
    if sys.platform == "darwin":
        dirs.insert(
            0, os.path.join(fsboot.executable_dir(), "..", "Resources")
        )

    locale_base = None
    for dir_ in dirs:
        check = os.path.join(dir_, "fs-uae-launcher", "share-dir")
        print("[I18N] Checking", check)
        if not os.path.exists(check):
            continue
        locale_base = os.path.join(dir_, "locale")
        break
    if not locale_base and getattr(sys, "frozen", False):
        if not locale_base:
            p = os.path.abspath(
                os.path.join(fsboot.executable_dir(), "..", "..", "Locale")
            )
            if os.path.exists(p):
                locale_base = p
        if not locale_base:
            p = os.path.abspath(
                os.path.join(
                    fsboot.executable_dir(), "..", "..", "Data", "Locale"
                )
            )
            if os.path.exists(p):
                locale_base = p
        if sys.platform == "darwin":
            # .app/Contents/Resources/Locale
            p = os.path.abspath(
                os.path.join(
                    fsboot.executable_dir(), "..", "Resources", "Locale"
                )
            )
            if os.path.exists(p):
                locale_base = p

    if locale_base:
        print("[I18N] bindtextdomain fs-uae-launcher:", locale_base)
        gettext_module.bindtextdomain("fs-uae-launcher", locale_base)

    mo_path = None
    if locale_base:
        print(
            "[I18N] find translations for",
            loc,
            "in local directory",
            locale_base,
        )
        try:
            mo_path = gettext_module.find(
                "fs-uae-launcher", locale_base, [loc]
            )
        except Exception as e:
            # a bug in openSUSE 12.2's gettext.py can cause an exception
            # in gettext.find (checking len of None).
            print(repr(e))
    else:
        print("[I18N]  No locale directory found")
    print("[I18N] Path to mo file:", mo_path)

    translations = gettext_module.translation(
        "fs-uae-launcher", locale_base, [loc], fallback=True
    )
    print("[I18N] Translations object:", translations)


def gettext(msg):
    if not _initialized:
        initialize_locale()
    return translations.gettext(msg)


def ngettext(n, msg1, msg2):
    if not _initialized:
        initialize_locale()
    return translations.ngettext(n, msg1, msg2)


translations = gettext_module.NullTranslations()
_ = gettext



import json

from fsgs import Option
from fsgs.drivers.mamedriver import MameDriver


class MameArcadeDriver(MameDriver):
    CONTROLLER = {
        "type": "controller",
        "description": "Controller",
        "mapping_name": "arcade",
    }

    PORTS = [
        {"description": "Player 1", "types": [CONTROLLER]},
        {"description": "Player 2", "types": [CONTROLLER]},
        {"description": "Player 3", "types": [CONTROLLER]},
        {"description": "Player 4", "types": [CONTROLLER]},
    ]

    def mame_romset(self):
        name = self.options["mame_rom_set"]
        files = {}
        for entry in json.loads(self.options["file_list"]):
            files[entry["name"]] = entry["sha1"]
        return name, files

    def mame_input_mapping(self, _):
        mapping = {
            "START": "START#",
            "SELECT": "COIN#",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
        }
        button_order = self.config.get(
            "button_order", "1 2 3 4 5 6 7 8 9 10 11 12"
        )
        b = 0
        for button in button_order.split(" "):
            button = button.strip()
            if button:
                b += 1
                mapping[str(b)] = "P#_BUTTON" + button
        return mapping

    def prepare(self):
        super().prepare()
        if self.options[Option.MAME_ARTWORK] == "1":
            pass
        else:
            self.create_mame_layout()



import os
import shutil
import hashlib

# TODO: review the algorithm and add support for saving information about
# (empty) directories.


class GameChangeHandler(object):
    def __init__(self, path):
        self._preserve_changes_dir = path

    def init(self, state_dir, ignore=[]):
        print("\n" + "-" * 79 + "\n" + "CHANGEHANDLER INIT")
        print(self._preserve_changes_dir)
        path = self._preserve_changes_dir
        if os.path.exists(state_dir):
            print("Merging preserved changes in", state_dir, "to", path)
            if state_dir[-1] == "/" or state_dir[-1] == "\\":
                state_dir = state_dir[:-1]
            lstate_dir = len(state_dir)
            first_iteration = True
            for dirpath, dir_names, file_names in os.walk(state_dir):
                for file_name in file_names:
                    ignore_file = False
                    if first_iteration:
                        for ig in ignore:
                            assert ig[0] == "*"
                            dummy, ext = os.path.splitext(file_name)
                            if ext.lower() == ig[1:]:
                                print("ignoring", file_name)
                                ignore_file = True
                                break
                    if ignore_file:
                        continue
                    sourcepath = os.path.join(dirpath, file_name)
                    destpath = os.path.join(path, sourcepath[lstate_dir + 1 :])
                    if os.path.getsize(sourcepath) == 17:
                        with open(sourcepath, "rb") as f:
                            if f.read() == "FILE_IS_DELETED":
                                print(
                                    "- removing file",
                                    sourcepath[lstate_dir + 1 :],
                                )
                                if os.path.exists(destpath):
                                    os.remove(destpath)
                                else:
                                    print("  (already gone)")
                                continue
                    print("- updating file", sourcepath[lstate_dir + 1 :])
                    if not os.path.isdir(os.path.dirname(destpath)):
                        os.makedirs(os.path.dirname(destpath))
                    shutil.copyfile(sourcepath, destpath)
                first_iteration = False
        else:
            print("no game state")
        self._preserve_changes_files = self.create_file_version_list(path)
        print("done")

    def update(self, state_dir):
        print("\n" + "-" * 79 + "\n" + "CHANGEHANDLER UPDATE")
        print("SRC", self._preserve_changes_dir)
        print("DST", state_dir)
        oldfiles = self._preserve_changes_files
        newfiles = self.create_file_version_list(self._preserve_changes_dir)
        print("checking files")
        for filename, newcs in newfiles.items():
            try:
                oldcs = oldfiles[filename]
            except KeyError:
                print("New file:", filename)
                oldcs = None
            if newcs != oldcs:
                print("File changed:", filename)
                print("-", newcs, "vs", oldcs)
                sourcepath = os.path.join(self._preserve_changes_dir, filename)
                destpath = os.path.join(state_dir, filename)
                print("Writing file", destpath)
                if not os.path.exists(os.path.dirname(destpath)):
                    os.makedirs(os.path.dirname(destpath))
                shutil.copyfile(sourcepath, destpath)

        for filename in oldfiles:
            if not filename in newfiles:
                print("File removed", filename)
                destpath = os.path.join(state_dir, filename)
                if not os.path.exists(os.path.dirname(destpath)):
                    os.makedirs(os.path.dirname(destpath))
                with open(destpath, "wb") as f:
                    f.write("FILE_IS_DELETED")
        print("done")

    def create_file_version_list(self, path):
        print("create_file_version_list")
        if path[-1] == "/" or path[-1] == "\\":
            path = path[:-1]
        lpath = len(path)
        files = {}
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                checksum = self.md5file(filepath)
                files[filepath[lpath + 1 :]] = checksum
        print(" - found %d files (checksummed)" % len(files))
        return files

    def md5file(self, file):
        with open(file, "rb") as f:
            m = hashlib.md5()
            while True:
                buffer = f.read(4096)
                if buffer == b"":
                    break
                m.update(buffer)
            return m.hexdigest()



from arcade.glui.opengl import gl, fs_emu_texturing, fs_emu_blending
from arcade.glui.text import TextRenderer

DISPLAY_FPS = False
DISPLAY_SYNC = False

PERSPECTIVE_STANDARD = 1
PERSPECTIVE_ORTHO = 2
PERSPECTIVE_HD = 3

TEXT_CACHE_SIZE = 100


class Render(object):
    _instance = None

    @classmethod
    def get(cls):
        if cls._instance is None:
            cls._instance = cls("DO_NOT_INSTANTIATE")
        return cls._instance

    @classmethod
    def reset(cls):
        print("[STATE] Reset state")
        cls._instance = None

    def __init__(self, sentinel):
        assert sentinel == "DO_NOT_INSTANTIATE"
        print("[STATE] Init state")

        self.was_scanning = False

        self.currently_ingame = False
        self.game_running = False

        # Render items
        self.display_sync = DISPLAY_SYNC
        self.display_fps = DISPLAY_FPS
        self.display_last_iteration = 0
        self.display_width = 0
        self.display_height = 0
        self.last_time_str = ""

        self.gl_left = None
        self.gl_right = None
        self.gl_height = None
        self.gl_width = None
        self.fov_y = None
        self.fov_x = None
        self.display_aspect = None
        self.ortho_pscalex = None
        self.ortho_pscaley = None
        self.projection = None

        # State items

        self.time = 0
        self.down_navigatable = None
        self.dialog = None
        self.dialog_time = 0.0

        # ...

        self.quit = False
        self.allow_minimize = True

        self.history = []

        self.zoom = 1.0
        self.offset_x = 0.0
        self.offset_y = 0.0

        self.frame_number = 0
        self.idle_from = None
        self.dirty = True
        self.non_dirty_state = False
        self.twice = False
        self.fade_start = 0.0
        self.fade_end = 0.0
        self.fade_splash = True
        self.fade_from = (0, 0, 0, 0)
        self.fade_to = (0, 0, 0, 0)
        self.opacity = 1.0
        self.perspective = 0

        self.center_item = None
        self.center_item_time = 0

        self.items_brightness = 0.8
        # self.items_brightness_target = 1.0
        # self.items_brightness_speed = 1.0
        self.navigatable = None
        self.current_menu = None
        self.current_game = None
        self.config_menu = None
        self.top_menu = None

        self.hide_mouse_time = 0
        self.mouse_visible = True
        # options
        self.force_portrait = False
        self.max_ratio = None
        # self.max_ratio = 1.0

        self.delete_texture_list = []
        self.unused_texture_list = []

        self.text_cache_history = []
        self.text_cache = {}

        self.mouse_item = None
        self.mouse_press_item = None

    def create_texture(self):
        if len(self.unused_texture_list) == 0:
            self.unused_texture_list.extend(gl.glGenTextures(50))
        return self.unused_texture_list.pop()

    def delete_textures(self):
        # for t in cls.delete_texture_list:
        #     glDeleteTextures([t])

        gl.glDeleteTextures(self.delete_texture_list)
        self.delete_texture_list[:] = []

        # cls.unused_texture_list.extend(cls.delete_texture_list)
        # cls.delete_texture_list[:] = []
        # pass

        if len(self.delete_texture_list) > 0:
            texture = self.delete_texture_list.pop()
            gl.glDeleteTextures(texture)

    def standard_perspective(self):
        if self.perspective == PERSPECTIVE_STANDARD:
            return
        # aspect_ratio = cls.display_width / cls.display_height
        self.gl_height = 2.0
        self.gl_width = 2.0 * self.display_aspect
        self.gl_left = -1.0 * self.display_aspect
        self.fov_y = 45 / self.zoom
        self.fov_x = self.fov_y * self.display_aspect / self.zoom

        # left = cls.offset_x - 1.0 * cls.display_aspect / cls.zoom
        # right = cls.offset_x + 1.0 * cls.display_aspect / cls.zoom
        # top = cls.offset_y + 1.0 / cls.zoom
        # bottom = cls.offset_y - 1.0 / cls.zoom

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()

        # gluPerspective(cls.fov_y, cls.display_aspect, 0.1, 10.1)
        # glTranslatef(-cls.offset_x, -cls.offset_y, 0.0)

        gl.gluPerspective(45.0, 16.0 / 9.0, 0.1, 100.0)

        self.projection = gl.glGetFloatv(gl.GL_PROJECTION_MATRIX)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        # glScissor(0, display_yoffset, cls.display_width, cls.display_height)
        # glEnable(GL_SCISSOR_TEST)
        self.perspective = PERSPECTIVE_STANDARD

    def hd_perspective(self):
        if self.perspective == PERSPECTIVE_HD:
            return
        self.display_aspect = 16 / 9.0
        self.ortho_pscalex = 1920 / self.display_width
        # cls.ortho_pscalex = 1920 / cls.display_width * \
        #         (cls.display_width / cls.display_height)
        self.ortho_pscaley = 1080 / self.display_height
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        left = 0.0
        right = 1920.0
        top = 1080.0
        bottom = 0.0
        gl.glOrtho(left, right, bottom, top, -1.0, 1.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.gl_left = left
        self.gl_right = right
        self.gl_height = 1080.0
        self.gl_width = 1920.0
        self.perspective = PERSPECTIVE_HD
        return self.ortho_pscalex, self.ortho_pscaley

    def ortho_perspective(self):
        if self.perspective == PERSPECTIVE_ORTHO:
            return
        self.display_aspect = 16 / 9.0
        self.ortho_pscalex = (
            2.0
            / self.display_width
            * (self.display_width / self.display_height)
        )
        self.ortho_pscaley = 2.0 / self.display_height
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        left = self.offset_x - 1.0 * self.display_aspect / self.zoom
        right = self.offset_x + 1.0 * self.display_aspect / self.zoom
        top = self.offset_y + 1.0 / self.zoom
        bottom = self.offset_y - 1.0 / self.zoom
        gl.glOrtho(left, right, bottom, top, -1.0, 1.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.gl_left = -1.0 * self.display_aspect
        self.gl_right = 1.0 * self.display_aspect
        self.gl_height = 2.0  # * cls.display_aspect
        self.gl_width = 2.0 * self.display_aspect
        self.perspective = PERSPECTIVE_ORTHO
        return self.ortho_pscalex, self.ortho_pscaley

    def text(
        self,
        text,
        font,
        x,
        y,
        w=0,
        h=0,
        color=(1.0, 1.0, 1.0, 1.0),
        shadow=False,
        shadow_color=(0, 0, 0),
        halign=-1,
    ):
        if not text:
            return 0, 0
        # if len(color) == 3:
        #     color = (color[0], color[1], color[2], 1.0
        try:
            alpha = color[3]
        except IndexError:
            alpha = 1.0
        color = (
            int(round(color[0] * 255)),
            int(round(color[1] * 255)),
            int(round(color[2] * 255)),
        )

        cache_key = (text, hash(font), font.size, color, alpha)
        try:
            self.text_cache_history.remove(cache_key)
        except ValueError:
            texture = None
        else:
            texture, txtsize = self.text_cache[cache_key]

        fs_emu_blending(True)
        fs_emu_texturing(True)
        gl.glDisable(gl.GL_DEPTH_TEST)

        if texture:
            gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
        else:
            txtdata, txtsize = TextRenderer(font).render_text(text, color)
            texture = Render.get().create_texture()
            gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
            gl.glTexImage2D(
                gl.GL_TEXTURE_2D,
                0,
                gl.GL_RGBA,
                txtsize[0],
                txtsize[1],
                0,
                gl.GL_BGRA,
                gl.GL_UNSIGNED_BYTE,
                txtdata,
            )
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR
            )
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR
            )
        tw, th = (
            txtsize[0] * self.ortho_pscalex,
            txtsize[1] * self.ortho_pscaley,
        )

        tx = x
        ty = y
        if w > 0:
            if tw > w:
                tw = w
            else:
                if halign == 0:
                    tx += (w - tw) / 2
        if h > 0:
            ty += (h - th) / 2
        # ts = 4 / cls.display_height  # Step

        # glDisable(GL_TEXTURE_RECTANGLE_ARB)

        # glTexEnv(GL_TEXTURE_2D, GL_MODULATE)
        # glDisable(GL_TEXTURE_RECTANGLE)
        # fs_emu_blending(True)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor4f(alpha, alpha, alpha, alpha)

        gl.glTexCoord2f(0.0, 0.0)
        gl.glVertex2f(tx, ty + th)
        gl.glTexCoord2f(1.0, 0.0)
        gl.glVertex2f(tx + tw, ty + th)
        gl.glTexCoord2f(1.0, 1.0)
        gl.glVertex2f(tx + tw, ty)
        gl.glTexCoord2f(0.0, 1.0)
        gl.glVertex2f(tx, ty)

        # glRasterPos2f(tx, ty)
        # glDrawPixels(txtsize[0], txtsize[1], GL_RGBA, GL_UNSIGNED_BYTE,
        # txtdata)
        gl.glEnd()

        # fs_emu_blending(False)
        gl.glEnable(gl.GL_DEPTH_TEST)

        self.text_cache_history.append(cache_key)
        self.text_cache[cache_key] = texture, txtsize
        if len(self.text_cache) > TEXT_CACHE_SIZE:
            cache_key = self.text_cache_history.pop(0)
            texture, txtsize = self.text_cache.pop(cache_key)
            Render.get().delete_texture_list.append(texture)

        # # FIXME:
        # shadow = False
        #
        # glDisable(GL_DEPTH_TEST)
        # fs_emu_blending(True)
        # #text = current_menu.selected_item.title
        # #if shadow:
        # txtdata, txtsize = TextRenderer(font).render_text(text, shadow_color)
        # tw, th = txtsize[0] * ortho_pscalex, txtsize[1] * ortho_pscaley
        # tx = x
        # ty = y
        # if w > 0:
        #     tx = tx + (w - tw) / 2
        # if h > 0:
        #     ty = ty + (h - th) / 2
        # #tx = 0 - tw / 2
        # #ty = -0.67
        # ts = 4 / State.get().display_height # Step
        # if shadow:
        #     glPixelTransferf(GL_ALPHA_SCALE, 0.04)
        #     for fx, fy in [(1, 1), (-1, -1), (1, -1), (-1, 1), (1, 0), (-1,
        #  0),
        #             (0, -1), (0, 1)]:
        #         glRasterPos2f(tx - fx * ts, ty - fy * ts)
        #         glDrawPixels(txtsize[0], txtsize[1], GL_RGBA,
        # GL_UNSIGNED_BYTE,
        #                 txtdata)
        #     glPixelTransferf(GL_ALPHA_SCALE, 0.01)
        #     for fx, fy in [(0, 2), (2, 0), (0, -2), (-2, 0),
        #             (1, 2), (2, 1), (-1, 2), (-2, 1), (1, -2),
        #             (2, -1), (-1, -2), (-2, -1)]:
        #         glRasterPos2f(tx - fx * ts, ty - fy * ts)
        #         glDrawPixels(txtsize[0], txtsize[1], GL_RGBA,
        # GL_UNSIGNED_BYTE,
        #                 txtdata)
        # glPixelTransferf(GL_ALPHA_SCALE, alpha)
        # rendered = font.render(text, True, color)
        # txtsize = rendered.get_size()
        # txtdata = pygame.image.tostring(rendered, "RGBA", 1)
        # glRasterPos2f(tx, ty)
        # glDrawPixels(txtsize[0], txtsize[1], GL_RGBA, GL_UNSIGNED_BYTE,
        # txtdata)
        # #glPopAttrib()
        # glPixelTransferf(GL_ALPHA_SCALE, 1.0)
        # fs_emu_blending(False)
        # glEnable(GL_DEPTH_TEST)
        return tw, th

    def measure_text(self, text, font):
        tw, th = TextRenderer(font).get_size(text)
        return tw * self.ortho_pscalex, th * self.ortho_pscaley


State = Render



import fsui
from fsui.extra.iconheader import NewIconHeader
from launcher.i18n import gettext
from launcher.settings.option_ui import OptionUI


class FilterSettingsPage(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self,
            fsui.Icon("video-settings", "pkg:workspace"),
            gettext("Scaling Options"),
            "",
        )
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        def add_option(name):
            self.layout.add(
                OptionUI.create_group(self, name),
                fill=True,
                margin_top=10,
                margin_bottom=10,
            )

        add_option("zoom")
        add_option("keep_aspect")

        label = fsui.HeadingLabel(self, gettext("Filters"))
        self.layout.add(label, margin_top=20, margin_bottom=20)

        add_option("scanlines")
        add_option("rtg_scanlines")
        # FIXME: gamma
        # FIXME: brightness / contrast / saturation



from RomAccess import RomAccess


RTC_MATCHWORD  = 0x4AFC

RTF_AUTOINIT   = (1<<7)
RTF_AFTERDOS   = (1<<2)
RTF_SINGLETASK = (1<<1)
RTF_COLDSTART  = (1<<0)

flag_names = {
  RTF_AUTOINIT: "RTF_AUTOINIT",
  RTF_AFTERDOS: "RTF_AFTERDOS",
  RTF_SINGLETASK: "RTF_SINGLETASK",
  RTF_COLDSTART: "RTF_COLDSTART"
}

NT_UNKNOWN = 0
NT_TASK = 1
NT_DEVICE = 3
NT_RESOURCE = 8
NT_LIBRARY = 9

nt_names = {
  NT_UNKNOWN: "NT_UNKNOWN",
  NT_TASK: "NT_TASK",
  NT_DEVICE: "NT_DEVICE",
  NT_RESOURCE: "NT_RESOURCE",
  NT_LIBRARY: "NT_LIBRARY"
}


class Resident:
  def __init__(self, off, flags, version, node_type, pri, name, id_string, init_off, skip_off):
    self.off = off
    self.flags = flags
    self.version = version
    self.node_type = node_type
    self.pri = pri
    self.name = name
    self.id_string = id_string
    self.init_off = init_off
    self.skip_off = skip_off

  def __repr__(self):
    return "Resident(@off=%08x,flags=%02x,version=%d,node_type=%d,pri=%d," \
           "name=%s,id_string=%s,init_off=%08x,skip_off=%08x)" % \
      (self.off, self.flags, self.version, self.node_type, self.pri,
       self.name, self.id_string, self.init_off, self.skip_off)

  def get_flags_strings(self):
    f = self.flags
    res = []
    for fn in flag_names:
      if f & fn == fn:
        res.append(flag_names[fn])
    return res

  def get_node_type_str(self):
    nt = self.node_type
    if nt in nt_names:
      return nt_names[nt]
    else:
      return str(nt)

  @classmethod
  def parse(cls, access, off, base_addr):
    # +0 RT_MATCHWORD
    mw = access.read_word(off)
    if mw != RTC_MATCHWORD:
      raise ValueError("No RTC_MATCHWORD at resident offset!")
    # +2 RT_MATCHTAG
    tag_ptr = access.read_long(off+2)
    if tag_ptr != base_addr + off:
      raise ValueError("Wrong MatchTag pointer in resident!")
    # +6 RT_ENDSKIP
    end_skip_ptr = access.read_long(off+6)
    end_skip_off = end_skip_ptr - base_addr
    # +10..13 RT_FLAGS, RT_VERSION, RT_TYPE, RT_PRI
    flags = access.read_byte(off+10)
    version = access.read_byte(off+11)
    rtype = access.read_byte(off+12)
    pri = access.read_sbyte(off+13)
    # +14: RT_NAME
    name = cls._parse_cstr(access, off+14, base_addr)
    # +18: RT_IDSTRING
    id_string = cls._parse_cstr(access, off+18, base_addr)
    # +22: RT_INIT
    init_ptr = access.read_long(off+22)
    init_off = init_ptr - base_addr
    return Resident(off, flags, version, rtype, pri, name, id_string,
                    init_off, end_skip_off)

  @classmethod
  def _parse_cstr(cls, access, off, base_addr):
    str_ptr = access.read_long(off)
    str_off = str_ptr - base_addr
    res = ""
    rom = access.rom_data
    while True:
      c = rom[str_off]
      if ord(c) == 0:
        break
      res += c
      str_off += 1
    return res


class ResidentScan:

  def __init__(self, rom_data, base_addr=0):
    self.access = RomAccess(rom_data)
    self.base_addr = base_addr

  def get_all_matchwords(self):
    """scan memory for all occurrences of matchwords"""
    mw = "\x4a\xfc"
    res = []
    off = 0
    num = self.access.size
    rom = self.access.rom_data
    while off < num:
      pos = rom.find(mw, off)
      if pos == -1:
        break
      res.append(pos)
      off = pos + 2
    return res

  def guess_base_addr(self):
    offs = self.get_all_matchwords()
    if len(offs) == 0:
      return None
    base_map = {}
    for off in offs:
      tag_ptr = self.access.read_long(off+2)
      tag_off = tag_ptr & 0xffff
      if tag_off == off:
        base_addr = tag_ptr & ~0xffff
        if base_addr not in base_map:
          base_map[base_addr] = 1
        else:
          base_map[base_addr] += 1
    # one match
    if len(base_map) == 1:
      addr = base_map.keys()[0]
      return addr
    else:
      return sorted(base_map.keys(), key=lambda x:base_map[x])

  def get_all_resident_pos(self):
    offs = self.get_all_matchwords()
    res = []
    for off in offs:
      # check tag ptr
      tag_ptr = self.access.read_long(off+2)
      if tag_ptr == self.base_addr + off:
        res.append(off)
    return res

  def is_resident_at(self, off):
    mw = self.access.read_word(off)
    if mw != RTC_MATCHWORD:
      return False
    tag_ptr = self.access.read_long(off+2)
    return tag_ptr == self.base_addr + off

  def get_resident(self, off):
    return Resident.parse(self.access, off, self.base_addr)
